import copy
import logging
import time
from ipaddress import IPv4Address
from pathlib import Path
from typing import Any, Callable, Collection, Dict, Final, Optional, Set, Tuple, Union, cast

import numpy as np
import numpy.typing as npt
from e7awghal import AwgParam, CapParam, E7FwType

from quel_ic_config.box_lock import BoxLockError
from quel_ic_config.e7resource_mapper import (
    AbstractQuel1E7ResourceMapper,
    create_rmap_object,
    validate_configuration_integrity,
)
from quel_ic_config.linkupper import LinkupFpgaMxfe
from quel_ic_config.quel1_any_config_subsystem import Quel1AnyConfigSubsystem
from quel_ic_config.quel1_config_loader import Quel1ConfigLoader
from quel_ic_config.quel1_config_subsystem import (
    QubeOuTypeAConfigSubsystem,
    QubeOuTypeBConfigSubsystem,
    Quel1BoxType,
    Quel1NecConfigSubsystem,
    Quel1TypeAConfigSubsystem,
    Quel1TypeBConfigSubsystem,
)
from quel_ic_config.quel1_config_subsystem_common import NoRfSwitchError
from quel_ic_config.quel1_wave_subsystem import (
    AbstractCancellableTaskWrapper,
    AbstractStartAwgunitsTask,
    CapIqDataReader,
    Quel1WaveSubsystem,
    StartAwgunitsNowTask,
    StartAwgunitsTimedTask,
    StartCapunitsByTriggerTask,
    StartCapunitsNowTask,
)
from quel_ic_config.quel1se_adda_config_subsystem import Quel1seAddaConfigSubsystem, Quel2ProtoAddaConfigSubsystem
from quel_ic_config.quel1se_fujitsu11_config_subsystem import (
    Quel1seFujitsu11TypeAConfigSubsystem,
    Quel1seFujitsu11TypeADebugConfigSubsystem,
    Quel1seFujitsu11TypeBConfigSubsystem,
    Quel1seFujitsu11TypeBDebugConfigSubsystem,
)
from quel_ic_config.quel1se_riken8_config_subsystem import (
    Quel1seRiken8ConfigSubsystem,
    Quel1seRiken8DebugConfigSubsystem,
)
from quel_ic_config.quel_config_common import Quel1ConfigOption, Quel1Feature

logger = logging.getLogger(__name__)

Quel1LineType = tuple[int, Union[int, str]]


def _complete_ipaddrs(ipaddr_wss: str, ipaddr_sss: Union[str, None], ipaddr_css: Union[str, None]) -> Tuple[str, str]:
    if ipaddr_sss is None:
        ipaddr_sss = str(IPv4Address(ipaddr_wss) + (1 << 16))
    if ipaddr_css is None:
        ipaddr_css = str(IPv4Address(ipaddr_wss) + (4 << 16))
    return ipaddr_sss, ipaddr_css


def _create_css_object(
    ipaddr_css: str,
    boxtype: Quel1BoxType,
) -> Quel1AnyConfigSubsystem:
    if boxtype in {Quel1BoxType.QuBE_RIKEN_TypeA, Quel1BoxType.QuEL1_TypeA}:
        css: Quel1AnyConfigSubsystem = Quel1TypeAConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype in {Quel1BoxType.QuBE_RIKEN_TypeB, Quel1BoxType.QuEL1_TypeB}:
        css = Quel1TypeBConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuBE_OU_TypeA:
        css = QubeOuTypeAConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuBE_OU_TypeB:
        css = QubeOuTypeBConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuEL1_NEC:
        css = Quel1NecConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuEL1SE_Adda:
        css = Quel1seAddaConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuEL2_ProtoAdda:
        css = Quel2ProtoAddaConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuEL1SE_RIKEN8:
        css = Quel1seRiken8ConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuEL1SE_FUJITSU11_TypeA:
        css = Quel1seFujitsu11TypeAConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuEL1SE_FUJITSU11_TypeB:
        css = Quel1seFujitsu11TypeBConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuEL1SE_RIKEN8DBG:
        css = Quel1seRiken8DebugConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuEL1SE_FUJITSU11DBG_TypeA:
        css = Quel1seFujitsu11TypeADebugConfigSubsystem(ipaddr_css, boxtype)
    elif boxtype == Quel1BoxType.QuEL1SE_FUJITSU11DBG_TypeB:
        css = Quel1seFujitsu11TypeBDebugConfigSubsystem(ipaddr_css, boxtype)
    else:
        raise ValueError(f"unsupported boxtype: {boxtype}")

    # TODO: should be moved to the right place.
    # Notes: for SockClients, this check is meaningless because lock acquisition is completed at the end of ctor.
    # Notes: for CoapClients, this looks required.
    for i in range(10):
        if i > 0:
            time.sleep(0.5)
        if css.has_lock:
            break
    else:
        del css
        raise BoxLockError(f"failed to acquire lock of {ipaddr_css}")

    return css  # noqa: F821 (avoiding a possible bug of pflake8)


def _dummy_auth_callback():
    # Notes: only for debug purpose. no access control will be applied.
    return True


def _create_wss_object(
    ipaddr_wss: str, ipaddr_sss: Optional[str] = None, css: Optional[Quel1AnyConfigSubsystem] = None
) -> Quel1WaveSubsystem:
    if css:
        auth_callback: Optional[Callable[[], bool]] = lambda: css.has_lock
    else:
        logger.warning(f"creating wss at {ipaddr_wss} without any access control.")
        auth_callback = _dummy_auth_callback

    return Quel1WaveSubsystem(ipaddr_wss, ipaddr_sss, auth_callback)


def create_css_wss_rmap(
    *,
    ipaddr_wss: str,
    ipaddr_sss: Union[str, None] = None,
    ipaddr_css: Union[str, None] = None,
    boxtype: Union[Quel1BoxType, str],
) -> tuple[Quel1AnyConfigSubsystem, Quel1WaveSubsystem, AbstractQuel1E7ResourceMapper]:
    ipaddr_sss, ipaddr_css = _complete_ipaddrs(ipaddr_wss, ipaddr_sss, ipaddr_css)
    if isinstance(boxtype, str):
        boxtype = Quel1BoxType.fromstr(boxtype)

    css: Quel1AnyConfigSubsystem = _create_css_object(ipaddr_css, boxtype)
    wss: Quel1WaveSubsystem = _create_wss_object(ipaddr_wss, ipaddr_sss, css)
    css.initialize()
    rmap: AbstractQuel1E7ResourceMapper = create_rmap_object(boxname=ipaddr_wss, fw_type=wss.fw_type)
    # Notes: wss will be initialized Boxi.initialize()
    return css, wss, rmap


class BoxIntrinsicStartCapunitsNowTask(
    AbstractCancellableTaskWrapper[dict[tuple[int, int], CapIqDataReader], dict[tuple[int, str, int], CapIqDataReader]]
):
    def __init__(self, task: StartCapunitsNowTask, mapping: dict[tuple[int, int], tuple[int, str, int]]):
        super().__init__(task)
        self._mapping = mapping

    def _conveter(self, orig: dict[tuple[int, int], CapIqDataReader]) -> dict[tuple[int, str, int], CapIqDataReader]:
        return {self._mapping[capunit]: reader for capunit, reader in orig.items()}


class BoxIntrinsicStartCapunitsByTriggerTask(
    AbstractCancellableTaskWrapper[dict[tuple[int, int], CapIqDataReader], dict[tuple[int, str, int], CapIqDataReader]]
):
    def __init__(self, task: StartCapunitsByTriggerTask, mapping: dict[tuple[int, int], tuple[int, str, int]]):
        super().__init__(task)
        self._mapping = mapping

    def _conveter(self, orig: dict[tuple[int, int], CapIqDataReader]) -> dict[tuple[int, str, int], CapIqDataReader]:
        return {self._mapping[capunit]: reader for capunit, reader in orig.items()}


class Quel1BoxIntrinsic:
    __slots__ = (
        "_css",
        "_wss",
        "_rmap",
        "_linkupper",
        "_options",
    )

    NUM_SAMPLE_IN_WAVE_BLOCK: Final[int] = 64  # TODO: should be defined in wss.
    NUM_SAMPLE_IN_WORD: Final[int] = 4  # TODO: should be defined in wss.
    DEFAULT_AMPLITUDE: Final[float] = 16383.0
    DEFAULT_NUM_WAVE_SAMPLE: Final[int] = NUM_SAMPLE_IN_WAVE_BLOCK * 1
    DEFAULT_REPEATS: Final[Tuple[int, int]] = (0xFFFFFFFF, 0xFFFFFFFF)
    DEFAULT_NUM_WAIT_SAMPLES: Final[Tuple[int, int]] = (0, 0)
    DEFAULT_NUM_CAPTURE_SAMPLE: Final[int] = 4096
    VERY_LONG_DURATION: float = 0x10000000000000000 * 128e-9
    DEFAULT_SCHEDULE_DEADLINE: Final[float] = 0.25  # [s]
    DEFAULT_SCHEDULE_WINDOW: Final[float] = 300.0  # [s]

    @classmethod
    def create(
        cls,
        *,
        ipaddr_wss: str,
        ipaddr_sss: Union[str, None] = None,
        ipaddr_css: Union[str, None] = None,
        boxtype: Quel1BoxType,
        skip_init: bool = False,
        **options: Collection[int],
    ) -> "Quel1BoxIntrinsic":
        """create QuEL intrinsic box objects
        :param ipaddr_wss: IP address of the wave generation subsystem of the target box
        :param ipaddr_sss: IP address of the sequencer subsystem of the target box (optional)
        :param ipaddr_css: IP address of the configuration subsystem of the target box (optional)
        :param boxtype: type of the target box
        :param config_root: root path of config setting files to read (optional)
        :param config_options: a collection of config options (optional)
        :param skip_init: skip calling box.initialization(), just for debugging.
        :param ignore_crc_error_of_mxfe: a list of MxFEs whose CRC error of the datalink is ignored. (optional)
        :param ignore_access_failure_of_adrf6780: a list of ADRF6780 whose communication faiulre via SPI bus is
                                                  dismissed (optional)
        :param ignore_lock_failure_of_lmx2594: a list of LMX2594 whose lock failure is ignored (optional)
        :param ignore_extraordinary_converter_select_of_mxfe: a list of MxFEs whose unusual converter mapping is
                                                              dismissed (optional)
        :return: SimpleBoxIntrinsic objects
        """
        css, wss, rmap = create_css_wss_rmap(
            ipaddr_wss=ipaddr_wss,
            ipaddr_sss=ipaddr_sss,
            ipaddr_css=ipaddr_css,
            boxtype=boxtype,
        )
        box = Quel1BoxIntrinsic(css=css, wss=wss, rmap=rmap, linkupper=None, **options)
        if not skip_init:
            box.initialize()
        return box

    # TODO: consider to re-locate to the right place
    def _validate_options(self, flags: Dict[str, Collection[int]]):
        num_ad9082 = self.css.get_num_ic("ad9082")
        num_adrf6780 = self.css.get_num_ic("adrf6780")
        num_lmx2594 = self.css.get_num_ic("lmx2594")

        for k, v in flags.items():
            if k == "ignore_crc_error_of_mxfe":
                if not all([0 <= u < num_ad9082 for u in v]):
                    raise ValueError(f"invalid index of mxfe is found in {k} (= {v})")
            elif k == "ignore_access_failure_of_adrf6780":
                if not all([0 <= u < num_adrf6780 for u in v]):
                    raise ValueError(f"invalid index of adrf6780 is found in {k} (= {v})")
            elif k == "ignore_lock_failure_of_lmx2594":
                if not all([0 <= u < num_lmx2594 for u in v]):
                    raise ValueError(f"invalid index of lmx2594 is found in {k} (= {v})")
            elif k == "ignore_extraordinary_converter_select_of_mxfe":
                if not all([0 <= u < num_ad9082 for u in v]):
                    raise ValueError(f"invalid index of ad9082 is found in {k} (= {v})")
            else:
                raise ValueError(f"invalid workaround options: {k}")

        for k in (
            "ignore_crc_error_of_mxfe",
            "ignore_access_failure_of_adrf6780",
            "ignore_lock_failure_of_lmx2594",
            "ignore_extraordinary_converter_select_of_mxfe",
        ):
            if k not in flags:
                flags[k] = set()

    def __init__(
        self,
        *,
        css: Quel1AnyConfigSubsystem,
        wss: Quel1WaveSubsystem,
        rmap: AbstractQuel1E7ResourceMapper,
        linkupper: Union[LinkupFpgaMxfe, None] = None,
        **options: Collection[int],
    ):
        self._css = css
        self._wss = wss
        self._rmap = rmap
        if linkupper is None:
            linkupper = LinkupFpgaMxfe(css, wss, rmap)
        self._linkupper = linkupper
        if options is None:
            options = {}
        self._validate_options(options)
        self._options: Dict[str, Collection[int]] = options

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}:{self._wss.ipaddr_wss}:{self.boxtype}>"

    @property
    def css(self) -> Quel1AnyConfigSubsystem:
        return self._css

    @property
    def wss(self) -> Quel1WaveSubsystem:
        return self._wss

    @property
    def rmap(self) -> AbstractQuel1E7ResourceMapper:
        return self._rmap

    @property
    def linkupper(self) -> LinkupFpgaMxfe:
        return self._linkupper

    @property
    def boxtype(self) -> str:
        return self.css.boxtype.tostr()

    @property
    def has_lock(self) -> bool:
        return self.css.has_lock

    @property
    def options(self) -> Dict[str, Collection[int]]:
        return self._options

    def initialize(self) -> None:
        self.wss.initialize()

    def reconnect(
        self,
        *,
        mxfe_list: Union[Collection[int], None] = None,
        skip_capture_check: bool = False,
        background_noise_threshold: Union[float, None] = None,
        ignore_crc_error_of_mxfe: Union[Collection[int], None] = None,
        ignore_extraordinary_converter_select_of_mxfe: Union[Collection[int], None] = None,
        ignore_invalid_linkstatus: bool = False,
    ) -> Dict[int, bool]:
        """establish a configuration link between a box and host.
        the target box needs to be linked-up in advance.

        :param mxfe_list: a list of MxFEs to reconnect. (optional)
        :param skip_capture_check: do not check background noise of input lines if True (optional)
        :param background_noise_threshold: the largest peak of allowable noise (optional)
        :param ignore_crc_error_of_mxfe: a list of MxFEs whose CRC error of the datalink is ignored. (optional)
        :param ignore_extraordinary_converter_select_of_mxfe: a list of MxFEs whose unusual converter mapping is
                                                              dismissed. (optional)
        :return: True if success
        """

        # Notes: boxi.initialize() is called at BoxIntrinsic.create() unless skip_init is True.
        #        this means that all AWG units and Capture units are initialized (and should not be working).
        #        reconnect() is usually called just after boxi.create() and it is reasonable to assume that all the
        #        components of WSS are quiet.
        #        theoretically, reconnect() can be called any time in the user's code, but do not initialize WSS here
        #        to avoid unwanted destruction of configuration settings.

        if mxfe_list is None:
            mxfe_list = self._css.get_all_mxfes()

        if ignore_crc_error_of_mxfe is None:
            ignore_crc_error_of_mxfe = self._options["ignore_crc_error_of_mxfe"]

        if ignore_extraordinary_converter_select_of_mxfe is None:
            ignore_extraordinary_converter_select_of_mxfe = self._options[
                "ignore_extraordinary_converter_select_of_mxfe"
            ]

        link_ok = {}
        for mxfe_idx in mxfe_list:
            try:
                valid_link: bool = self._css.reconnect_mxfe(
                    mxfe_idx, ignore_crc_error=mxfe_idx in ignore_crc_error_of_mxfe
                )
                if valid_link:
                    self._css.validate_chip_id(mxfe_idx)
                    validate_configuration_integrity(
                        self.css.get_virtual_adc_select(mxfe_idx),
                        self.wss.fw_type,
                        ignore_extraordinary_converter_select=mxfe_idx in ignore_extraordinary_converter_select_of_mxfe,
                    )
            except RuntimeError as e:
                logger.warning(e)
                valid_link = False

            if not valid_link:
                if not ignore_invalid_linkstatus:
                    logger.error(f"JESD204C link of AD9082-#{mxfe_idx} is not working. it must be linked up")
            link_ok[mxfe_idx] = valid_link

        # Notes: checking so-called chikuchiku problem caused by wss firmware bug.
        if not skip_capture_check and all(link_ok.values()):
            rfsw_restore = {k: "open" for k, v in self.dump_rfswitches().items() if v == "open"}
            rfsw_closed = {k: "loop" for k in rfsw_restore}
            try:
                self.config_rfswitches(rfsw_closed)
            except NoRfSwitchError:
                logger.info(
                    "background noise check may be disrupted by incoming signal "
                    "due to the unavailability of RF switches"
                )
            for mxfe_idx in mxfe_list:
                if not self.linkupper.check_all_fddcs_of_mxfe_at_reconnect(mxfe_idx, background_noise_threshold):
                    link_ok[mxfe_idx] = False
                    if not ignore_invalid_linkstatus:
                        logger.error(
                            f"JESD204C tx-link of AD9082-#{mxfe_idx} is not working properly, it must be linked up"
                        )
                    # Notes: do not break here even if here to check all the MxFEs to show information.
            self.config_rfswitches(rfsw_restore)

        return link_ok

    def get_wss_features(self) -> set[Quel1Feature]:
        features: set[Quel1Feature] = set()
        if self.wss.fw_type in {E7FwType.SIMPLEMULTI_CLASSIC}:
            features.add(Quel1Feature.SINGLE_ADC)
        elif self.wss.fw_type in {E7FwType.FEEDBACK_VERYEARLY}:
            features.add(Quel1Feature.BOTH_ADC_EARLY)
        elif self.wss.fw_type in {E7FwType.SIMPLEMULTI_STANDARD, E7FwType.FEEDBACK_EARLY}:
            features.add(Quel1Feature.BOTH_ADC)
        else:
            raise ValueError(f"unsupported firmware is detected: {self.wss.fw_type}")
        return features

    def _load_config_parameter(
        self,
        *,
        config_root: Union[Path, None] = None,
        config_options: Union[Collection[Quel1ConfigOption], None] = None,
    ) -> Dict[str, Any]:
        config_options = config_options or set()
        loader = Quel1ConfigLoader(
            boxtype=self.css.boxtype,
            num_ic=self.css.get_num_ics(),
            config_options=config_options,
            features=self.get_wss_features(),
            config_filename=self.css.get_default_config_filename(),
            config_rootpath=config_root,
        )
        return loader.load_config()

    def hardreset_wss_units(
        self,
        mxfes: Union[Collection[int], None] = None,
        reset_awg: bool = True,
        reset_cap: bool = True,
        suppress_warning: bool = False,
    ) -> None:
        # Notes: an only way to clear hardware error flag of wss is hard-resetting the corresponding units.
        #        however, its soundness is not confirmed well yet. at least, it often destroys JESD204C link and needs
        #        the re-linkup. it is still unclear that re-linkup recovers the internal consistency again or not.
        #        be careful if you use it.
        if mxfes is None:
            mxfes = self._css.get_all_mxfes()

        if reset_awg:
            for mxfe in mxfes:
                for awg_idx in self._rmap.get_awgs_of_mxfe(mxfe):
                    self._wss._get_awgunit(awg_idx).hard_reset(suppress_warning=suppress_warning)

        if reset_cap:
            fddcs: Set[Tuple[int, int]] = set()
            for g in self._css.get_all_groups():
                for rl in self._css.get_all_rlines_of_group(g):
                    for rch in range(self._css.get_num_rchannels_of_rline(g, rl)):
                        m, d = self._css.get_fddc_idx(g, rl, rch)
                        fddcs.add((m, d))

            for mxfe, fddc in fddcs:
                if mxfe in mxfes:
                    cm_idx = self._rmap.get_capmod_from_fddc(mxfe, fddc)
                    for cu_idx in range(self._wss.num_capunit_of_capmod(cm_idx)):
                        self._wss._get_capunit((cm_idx, cu_idx)).hard_reset()

        if reset_awg or reset_cap:
            # TODO: confirm the effectiveness.
            logger.info("waiting for 15 seconds after hard-resetting wave subsystem")
            time.sleep(15)

    def relinkup(
        self,
        *,
        param: Union[Dict[str, Any], None] = None,
        config_root: Union[Path, None] = None,
        config_options: Union[Collection[Quel1ConfigOption], None] = None,
        mxfes_to_linkup: Union[Collection[int], None] = None,
        hard_reset: Union[bool, None] = None,
        use_204b: bool = False,
        use_bg_cal: bool = True,
        skip_init: bool = False,
        hard_reset_wss: bool = False,
        background_noise_threshold: Union[float, None] = None,
        ignore_crc_error_of_mxfe: Union[Collection[int], None] = None,
        ignore_access_failure_of_adrf6780: Union[Collection[int], None] = None,
        ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None,
        ignore_extraordinary_converter_select_of_mxfe: Union[Collection[int], None] = None,
        restart_tempctrl: bool = False,
    ) -> Dict[int, bool]:
        if mxfes_to_linkup is None:
            mxfes_to_linkup = self._css.get_all_mxfes()
        if hard_reset is None:
            hard_reset = self.css.boxtype.is_quel1se()
        if ignore_crc_error_of_mxfe is None:
            ignore_crc_error_of_mxfe = self._options["ignore_crc_error_of_mxfe"]
        if ignore_access_failure_of_adrf6780 is None:
            ignore_access_failure_of_adrf6780 = self._options["ignore_access_failure_of_adrf6780"]
        if ignore_lock_failure_of_lmx2594 is None:
            ignore_lock_failure_of_lmx2594 = self._options["ignore_lock_failure_of_lmx2594"]
        if ignore_extraordinary_converter_select_of_mxfe is None:
            ignore_extraordinary_converter_select_of_mxfe = self._options[
                "ignore_extraordinary_converter_select_of_mxfe"
            ]

        # Notes: user can inject a parameter object directly instead of loading it from files.
        param = param or self._load_config_parameter(config_root=config_root, config_options=config_options)

        if not skip_init:
            self.css.configure_peripherals(
                param,
                ignore_access_failure_of_adrf6780=ignore_access_failure_of_adrf6780,
                ignore_lock_failure_of_lmx2594=ignore_lock_failure_of_lmx2594,
            )
            self.css.configure_all_mxfe_clocks(
                param,
                ignore_lock_failure_of_lmx2594=ignore_lock_failure_of_lmx2594,
            )

        if hard_reset_wss:
            self.hardreset_wss_units(mxfes_to_linkup, suppress_warning=True)

        linkup_ok: Dict[int, bool] = {}
        for mxfe in mxfes_to_linkup:
            linkup_ok[mxfe] = self.linkupper.linkup_and_check(
                mxfe,
                param,
                hard_reset=hard_reset,
                use_204b=use_204b,
                use_bg_cal=use_bg_cal,
                background_noise_threshold=background_noise_threshold,
                ignore_crc_error=mxfe in ignore_crc_error_of_mxfe,
                ignore_extraordinary_converter_select=mxfe in ignore_extraordinary_converter_select_of_mxfe,
                restart_tempctrl=restart_tempctrl,
            )
        return linkup_ok

    def link_status(self, *, ignore_crc_error_of_mxfe: Union[Collection[int], None] = None) -> Dict[int, bool]:
        if ignore_crc_error_of_mxfe is None:
            ignore_crc_error_of_mxfe = self._options["ignore_crc_error_of_mxfe"]

        link_health: Dict[int, bool] = {}
        for mxfe_idx in self.css.get_all_mxfes():
            link_health[mxfe_idx], log_level, diag = self.css.check_link_status(
                mxfe_idx, False, ignore_crc_error=(mxfe_idx in ignore_crc_error_of_mxfe)
            )
            logger.log(log_level, diag)

        return link_health

    def _get_rchannel_from_runit(self, group: int, rline: str, runit: int) -> int:
        # TODO: implement a method to connect runit and rchannel in this (BoxIntrinsic) layer.
        if runit in self.get_runits_of_rline(group, rline):
            return 0
        else:
            raise ValueError(f"invalid runit:{runit} for group:{group}, rline:{rline}")

    def _get_awg_from_channel(self, group: int, line: int, channel: int) -> int:
        mxfe_idx, fduc_idx = self.css.get_fduc_idx(group, line, channel)
        return self.rmap.get_awg_from_fduc(mxfe_idx, fduc_idx)

    def _get_awgs_from_channels(self, channels: Collection[Tuple[int, int, int]]):
        return {self._get_awg_from_channel(gr, ln, ch) for gr, ln, ch in channels}

    def _get_capmod_from_rchannel(self, group: int, rline: str, rchannel: int) -> int:
        mxfe_idx, fddc_idx = self.css.get_fddc_idx(group, rline, rchannel)
        return self.rmap.get_capmod_from_fddc(mxfe_idx, fddc_idx)

    def _get_capunit_from_runit(self, group: int, rline: str, runit: int) -> tuple[int, int]:
        rchannel = self._get_rchannel_from_runit(group, rline, runit)
        # Notes: runit --> capunit doens't work in near future!
        return (self._get_capmod_from_rchannel(group, rline, rchannel), runit)

    def _get_capunits_from_runits(self, runits: Collection[tuple[int, str, int]]) -> set[tuple[int, int]]:
        return {self._get_capunit_from_runit(*runit) for runit in runits}

    def _config_box_inner_line(
        self,
        group: int,
        line: int,
        direction: Union[str, None],
        line_conf: Dict[str, Any],
        channel_confs: Dict[int, Dict[str, Any]],
        runit_confs: Dict[int, Dict[str, Any]],
    ):
        if runit_confs != {}:
            raise ValueError(f"'runits' is not applicable to an output line: ({group}, {line})")
        for ch, channel_conf in channel_confs.items():
            self.config_channel(group, line, ch, **channel_conf)
        self.config_line(group, line, **line_conf)
        if direction is not None:
            if direction == "out":
                logger.info(f"direction of ({group}, {line}) is confirmed to be {direction}")
            else:
                raise ValueError(
                    f"given direction '{direction}' doesn't match with the actual direction of ({group}, {line})"
                )

    def _config_box_inner_rline(
        self,
        group: int,
        rline: str,
        direction: Union[str, None],
        line_conf: Dict[str, Any],
        channel_confs: Dict[int, Dict[str, Any]],
        runit_confs: Dict[int, Dict[str, Any]],
    ):
        if channel_confs != {}:
            raise ValueError(f"'channels' is not applicable to an input line: ({group}, {rline})")
        for runit, runit_conf in runit_confs.items():
            self.config_runit(group, rline, runit, **runit_conf)
        self.config_rline(group, rline, **line_conf)
        if direction is not None:
            if direction == "in":
                logger.info(f"direction of ({group}, {rline}) is confirmed to be {direction}")
            else:
                raise ValueError(
                    f"given direction '{direction}' doesn't match with the actual direction of ({group}, {rline})"
                )

    def _config_box_inner(self, group: int, line: Union[int, str], lc: Dict[str, Any]):
        line_conf = copy.deepcopy(lc)
        if "direction" in line_conf:
            direction: Union[str, None] = line_conf["direction"]
            del line_conf["direction"]
        else:
            direction = None
        if "channels" in line_conf:
            channel_confs: Dict[int, Dict[str, Any]] = line_conf["channels"]
            del line_conf["channels"]
        else:
            channel_confs = {}
        if "runits" in line_conf:
            runit_confs: Dict[int, Dict[str, Any]] = line_conf["runits"]
            del line_conf["runits"]
        else:
            runit_confs = {}

        if self.is_output_line(group, line):
            self._config_box_inner_line(group, cast(int, line), direction, line_conf, channel_confs, runit_confs)
        elif self.is_input_line(group, line):
            self._config_box_inner_rline(group, cast(str, line), direction, line_conf, channel_confs, runit_confs)
        else:
            raise ValueError(f"invalid line: '{line}'")

    def config_box(self, box_conf: Dict[Tuple[int, Union[int, str]], Dict[str, Any]], ignore_validation: bool = False):
        # Notes: configure output lines before input ones to keep "cnco_locked_with" intuitive in config_box().
        for (group, line), lc in box_conf.items():
            if self.is_output_line(group, line):
                self._config_box_inner(group, line, lc)

        for (group, line), lc in box_conf.items():
            if self.is_input_line(group, line):
                self._config_box_inner(group, line, lc)

        if not ignore_validation:
            if not self.config_validate_box(box_conf):
                raise ValueError("the provided settings looks have inconsistent settings")

    def _config_validate_channels(
        self, group: int, line: Union[int, str], confs: Dict[int, Dict[str, Any]], parent_name: str
    ) -> bool:
        if not self.is_output_line(group, line):
            logger.error(f"'channels' appears at an unexpected {parent_name}")
            return False

        line = cast(int, line)
        valid = True
        for ch, cc in confs.items():
            acc = self.css.dump_channel(group, line, ch)
            for kk in cc:
                if kk not in acc:
                    valid = False
                    logger.error(f"unexpected settings at {parent_name}:channel-{ch}, '{kk}' is unavailable")
                elif cc[kk] != acc[kk]:
                    valid = False
                    logger.error(f"unexpected settings at {parent_name}:channel-{ch}: {kk} = {acc[kk]} (!= {cc[kk]})")
        return valid

    def _config_validate_runits(
        self, group: int, rline: Union[int, str], runits_conf: Dict[int, Dict[str, Any]], parent_name: str
    ) -> bool:
        if not self.is_input_line(group, rline):
            logger.error(f"'runits' is unavailable at {parent_name}")
            return False

        rline = cast(str, rline)
        valid = True
        for runit, uc in runits_conf.items():
            auc = self._dump_runit(group, rline, runit)
            for kk in uc:
                if kk not in auc:
                    valid = False
                    logger.error(f"unexpected settings at {parent_name}:runit-{runit}, '{kk}' is unavailable")
                elif uc[kk] != auc[kk]:
                    valid = False
                    logger.error(f"unexpected settings at {parent_name}:runit-{runit}: {kk} = {auc[kk]} (!= {uc[kk]})")
        return valid

    def _config_validate_frequency(self, group: int, line: Union[int, str], k: str, freq0: float, freq1: float) -> bool:
        if self.is_output_line(group, line):
            if k == "cnco_freq":
                return self.css.is_equivalent_dac_cnco(group, cast(int, line), freq0, freq1)
            elif k == "fnco_freq":
                return self.css.is_equivalent_dac_fnco(group, cast(int, line), freq0, freq1)
        elif self.is_input_line(group, line):
            if k == "cnco_freq":
                return self.css.is_equivalent_adc_cnco(group, cast(str, line), freq0, freq1)
            elif k == "fnco_freq":
                return self.css.is_equivalent_adc_fnco(group, cast(str, line), freq0, freq1)
        raise AssertionError

    def _config_validate_fsc(self, group: int, line: int, fsc0: int, fsc1: int) -> bool:
        return self._css.is_equal_fullscale_current(group, line, fsc0, fsc1)

    def _config_validate_line(self, group: int, line: Union[int, str], lc: Dict[str, Any]) -> bool:
        if self.is_output_line(group, line):
            line_name: str = f"group:{group}, line:{line}"
            alc: Dict[str, Any] = self.css.dump_line(group, cast(int, line))
            ad: str = "out"
        elif self.is_input_line(group, line):
            line_name = f"group:{group}, rline:{line}"
            alc = self.css.dump_rline(group, cast(str, line))
            ad = "in"
        else:
            raise ValueError(f"invalid line: ({group}, {line})")

        valid = True
        for k in lc:
            if k == "channels":
                if not self._config_validate_channels(group, line, lc["channels"], line_name):
                    valid = False
            elif k == "runits":
                if not self._config_validate_runits(group, line, lc["runits"], line_name):
                    valid = False
            elif k == "direction":
                if lc["direction"] != ad:
                    valid = False
                    logger.error(f"unexpected settings of {line_name}:" f"direction = {ad} (!= {lc['direction']})")
            elif k in {"cnco_freq", "fnco_freq"}:
                if not self._config_validate_frequency(group, line, k, lc[k], alc[k]):
                    valid = False
                    logger.error(f"unexpected settings at {line_name}:{k} = {alc[k]} (!= {lc[k]})")
            elif k == "cnco_locked_with":
                dac_g, dac_l = lc[k]
                alf = alc["cnco_freq"]
                lf = self._css.get_dac_cnco(dac_g, dac_l)
                if lf != alf:
                    valid = False
                    logger.error(
                        f"unexpected settings at {line_name}:cnco_freq = {alf} (!= {lf}, "
                        f"that is cnco frequency of group-{dac_g}:line-{dac_l})"
                    )
            elif k == "fullscale_current":
                if not self._config_validate_fsc(group, cast(int, line), lc[k], alc[k]):
                    valid = False
                    logger.error(f"unexpected settings at {line_name}:{k} = {alc[k]} (!= {lc[k]})")
            else:
                if lc[k] != alc[k]:
                    valid = False
                    logger.error(f"unexpected settings at {line_name}:{k} = {alc[k]} (!= {lc[k]})")
        return valid

    def config_validate_box(self, box_conf: Dict[Tuple[int, Union[int, str]], Dict[str, Any]]) -> bool:
        valid: bool = True
        for (group, line), lc in box_conf.items():
            valid &= self._config_validate_line(group, line, lc)
        return valid

    def is_output_line(self, group: int, line: Union[int, str]):
        return (group in self.css.get_all_groups()) and isinstance(line, int)

    def is_output_channel(self, group: int, line: Union[int, str], channel: int):
        return (
            (group in self.css.get_all_groups())
            and isinstance(line, int)
            and (channel in self.get_channels_of_line(group, line))
        )

    def is_input_line(self, group: int, line: Union[int, str]):
        return (group in self.css.get_all_groups()) and isinstance(line, str)

    def is_input_runit(self, group: int, rline: Union[int, str], runit: int):
        return (
            (group in self.css.get_all_groups())
            and isinstance(rline, str)
            and (runit in self.get_runits_of_rline(group, rline))
        )

    def is_read_input_line(self, group: int, line: Union[int, str]):
        return line == "r"

    def is_monitor_input_line(self, group: int, line: Union[int, str]):
        return line == "m"

    def config_line(
        self,
        group: int,
        line: int,
        *,
        lo_freq: Union[float, None] = None,
        cnco_freq: Union[float, None] = None,
        vatt: Union[int, None] = None,
        sideband: Union[str, None] = None,
        fullscale_current: Union[int, None] = None,
        rfswitch: Union[str, None] = None,
    ) -> None:
        """configuring parameters of a given transmitter line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :param lo_freq: frequency of the corresponding local oscillator in Hz, must be multiple of 100_000_000.
        :param cnco_freq: frequency of the corresponding CNCO in Hz.
        :param vatt: controlling voltage of the corresponding VATT in unit of 3.3V / 4096. see the specification
                     sheet of the ADRF6780 for details.
        :param sideband: "U" for upper side band, "L" for lower side band.
        :param fullscale_current: full-scale current of output DAC of AD9082 in uA.
        :param rfswitch: state of RF switch, 'block' or 'pass' for output line, "loop" or "open" for input line
        :return: None
        """
        if vatt is not None:
            self._css.set_vatt(group, line, vatt)
        if sideband is not None:
            self._css.set_sideband(group, line, sideband)
        if fullscale_current is not None:
            self._css.set_fullscale_current(group, line, fullscale_current)
        if lo_freq is not None:
            if 15_000_000_000 >= lo_freq >= 7_500_000_000:
                if lo_freq % 100000000 != 0:
                    raise ValueError("lo_freq must be multiple of 100000000 if 7.5GHz <= lo_freq <= 15GHz")
                self._css.set_divider_ratio(group, line, 1)
                self._css.set_lo_multiplier(group, line, int(lo_freq) // 100000000)
            elif lo_freq >= 3_750_000_000:
                if lo_freq % 50000000 != 0:
                    raise ValueError("lo_freq must be multiple of 50000000 if 3.75GHz <= lo_freq < 7.5GHz")
                self._css.set_divider_ratio(group, line, 2)
                self._css.set_lo_multiplier(group, line, int(lo_freq) // 50000000)
            elif lo_freq >= 1_875_000_000:
                if lo_freq % 25000000 != 0:
                    raise ValueError("lo_freq must be multiple of 25000000 if 1.875GHz <= lo_freq < 3.75GHz")
                self._css.set_divider_ratio(group, line, 4)
                self._css.set_lo_multiplier(group, line, int(lo_freq) // 25000000)
            else:
                raise ValueError(f"invalid lo_freq: {lo_freq}Hz")
        if cnco_freq is not None:
            self._css.set_dac_cnco(group, line, cnco_freq)
        if rfswitch is not None:
            self.config_rfswitch(group, line, rfswitch=rfswitch)

    def config_channel(
        self,
        group: int,
        line: int,
        channel: int,
        *,
        fnco_freq: Union[float, None] = None,
        awg_param: Union[AwgParam, None] = None,
    ) -> None:
        """configuring parameters of a given transmitter channel.

        :param group: a group which the channel belongs to.
        :param line: a group-local index of a line which the channel belongs to.
        :param channel: a line-local index of the channel.
        :param fnco_freq: the frequency of the corresponding FNCO in Hz. it must be within the range of -250e6 and
                          250e6.
        :param awg_param: an object holding parameters of signal to be generated.
        :return: None
        """
        if fnco_freq is not None:
            self._css.set_dac_fnco(group, line, channel, fnco_freq)

        if awg_param is not None:
            self._wss.config_awgunit(self._get_awg_from_channel(group, line, channel), awg_param)

    def config_rline(
        self,
        group: int,
        rline: str,
        *,
        lo_freq: Union[float, None] = None,
        cnco_freq: Union[float, None] = None,
        cnco_locked_with: Union[Tuple[int, int], None] = None,
        rfswitch: Union[str, None] = None,
    ) -> None:
        """configuring parameters of a given receiver line.

        :param group: an index of a group which the line belongs to.
        :param rline: a group-local index of the line.
        :param lo_freq: frequency of the corresponding local oscillator in Hz, must be multiple of 100_000_000.
        :param cnco_freq: frequency of the corresponding CNCO in Hz.
        :param cnco_locked_with: frequency of CNCO is constrained to be identical to the specified line.
        :param rfswitch: state of RF switch, 'loop' or 'open'
        :return: None
        """
        if lo_freq is not None:
            if 15_000_000_000 >= lo_freq >= 7_500_000_000:
                if lo_freq % 100000000 != 0:
                    raise ValueError("lo_freq must be multiple of 100000000 if 7.5GHz <= lo_freq <= 15GHz")
                self._css.set_divider_ratio(group, rline, 1)
                self._css.set_lo_multiplier(group, rline, int(lo_freq) // 100000000)
            elif lo_freq >= 3_750_000_000:
                if lo_freq % 50000000 != 0:
                    raise ValueError("lo_freq must be multiple of 50000000 if 3.75GHz <= lo_freq < 7.5GHz")
                self._css.set_divider_ratio(group, rline, 2)
                self._css.set_lo_multiplier(group, rline, int(lo_freq) // 50000000)
            elif lo_freq >= 1_875_000_000:
                if lo_freq % 25000000 != 0:
                    raise ValueError("lo_freq must be multiple of 25000000 if 1.875GHz <= lo_freq < 3.75GHz")
                self._css.set_divider_ratio(group, rline, 4)
                self._css.set_lo_multiplier(group, rline, int(lo_freq) // 25000000)
            else:
                raise ValueError(f"invalid lo_freq: {lo_freq}Hz")
        if cnco_freq is not None and cnco_locked_with is not None:
            raise ValueError("it is not allowed to specify both cnco_freq and cnco_locked_with at the same time")
        if cnco_freq is not None:
            self._css.set_adc_cnco(group, rline, freq_in_hz=cnco_freq)
        if cnco_locked_with is not None:
            if isinstance(cnco_locked_with, tuple) and len(cnco_locked_with) == 2:
                dac_group, dac_line = cnco_locked_with
                self._css.set_pair_cnco(dac_group, dac_line, group, rline, self._css.get_dac_cnco(dac_group, dac_line))
            else:
                raise ValueError(f"invalid line to locked with: {cnco_locked_with}")
        if rfswitch is not None:
            self.config_rfswitch(group, rline, rfswitch=rfswitch)

    def config_runit(
        self,
        group: int,
        rline: str,
        runit: int,
        *,
        fnco_freq: Union[float, None] = None,
        capture_param: Union[CapParam, None] = None,
    ) -> None:
        """configuring parameters of a given receiver channel.

        :param group: an index of a group which the channel belongs to.
        :param rline: a group-local index of a line which the channel belongs to.
        :param runit: a line-local index of the capture unit.
        :param fnco_freq: the frequency of the corresponding FNCO in Hz. it must be within the range of -250e6 and
                          250e6.
        :param capture_param: an object holding capture settings.
        :return: None
        """
        if fnco_freq is not None:
            rchannel = self._get_rchannel_from_runit(group, rline, runit)
            self._css.set_adc_fnco(group, rline, rchannel, freq_in_hz=fnco_freq)

        if capture_param is not None:
            capunit = self._get_capunit_from_runit(group, rline, runit)
            self._wss.config_capunit(capunit, capture_param)

    def block_all_output_lines(self) -> None:
        """set RF switch of all output lines to block.

        :return:
        """
        for group in self._css.get_all_groups():
            for line in self._css.get_all_lines_of_group(group):
                self.config_rfswitch(group, line, rfswitch="block")

    def pass_all_output_lines(self):
        """set RF switch of all output lines to pass.

        :return:
        """
        for group in self._css.get_all_groups():
            for line in self._css.get_all_lines_of_group(group):
                self.config_rfswitch(group, line, rfswitch="pass")

    def config_rfswitches(self, rfswich_confs: Dict[Tuple[int, Union[int, str]], str], ignore_validation=False) -> None:
        """configuring multiple RF switches at the same time. contradictions among the given configuration is checked.

        :param rfswich_confs: a mapping between a line and the configuration of its RF switch.
        :param ignore_validation: raises exception the given configuration is not fulfilled if True (default).
        :return:
        """
        for (group, line), rc in rfswich_confs.items():
            self.config_rfswitch(group, line, rfswitch=rc)

        valid = True
        for (group, line), rc in rfswich_confs.items():
            arc = self.dump_rfswitch(group, line)
            if arc != rc:
                valid = False
                logger.warning(f"rfswitch[{group}, {line}] is finally set to {arc} (= {rc})")
        if not (ignore_validation or valid):
            raise ValueError("the specified configuration of rf switches is not realizable")

    def config_rfswitch(
        self, group: int, line: Union[int, str], *, rfswitch: str, reconfig_forcibly: bool = False
    ) -> None:
        """configuring a single RF switch of a given line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :param rfswitch: "block" or "pass" for an output line. "loop" or "open" for an input line.
        :return:
        """
        cc = self.dump_rfswitch(group, line)
        if reconfig_forcibly or (cc != rfswitch):
            if self._encode_rfswitch_conf(group, line, rfswitch):
                self.close_rfswitch(group, line)
            else:
                self.open_rfswitch(group, line)

    def open_rfswitch(self, group: int, line: Union[int, str]):
        """opening RF switch of the port corresponding to a given line, either of transmitter or receiver one.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        return None
        """
        self._css.pass_line(group, line)

    def close_rfswitch(self, group: int, line: Union[int, str]):
        """closing RF switch of the port corresponding to a given line, either of transmitter or receiver one.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        return None
        """
        self._css.block_line(group, line)

    def _decode_rfswitch_conf(self, group: int, line: Union[int, str], block: bool) -> str:
        if self.is_output_line(group, line):
            return "block" if block else "pass"
        elif self.is_input_line(group, line):
            return "loop" if block else "open"
        else:
            raise ValueError(f"invalid group:{group}, line:{line})")

    def _encode_rfswitch_conf(self, group: int, line: Union[int, str], conf: str) -> bool:
        if self.is_output_line(group, line):
            if conf == "block":
                return True
            elif conf == "pass":
                return False
            else:
                raise ValueError(f"invalid configuration of an output switch: {conf}")
        elif self.is_input_line(group, line):
            if conf == "loop":
                return True
            elif conf == "open":
                return False
            else:
                raise ValueError(f"invalid configuration of an input switch: {conf}")
        else:
            raise ValueError(f"invalid group:{group}, line:{line}")

    def activate_monitor_loop(self, group: int) -> None:
        """enabling an internal monitor loop-back path from a monitor-out port to a monitor-in port.

        :param group: an index of a group which the monitor path belongs to.
        :return: None
        """
        self._css.activate_monitor_loop(group)

    def deactivate_monitor_loop(self, group: int) -> None:
        """disabling an internal monitor loop-back path.

        :param group: a group which the monitor path belongs to.
        :return: None
        """
        self._css.deactivate_monitor_loop(group)

    def is_loopedback_monitor(self, group: int) -> bool:
        """checking if an internal monitor loop-back path is activated or not.

        :param group: an index of a group which the monitor loop-back path belongs to.
        :return: True if the monitor loop-back path is activated.
        """
        return self._css.is_loopedback_monitor(group)

    def activate_read_loop(self, group: int) -> None:
        """enabling an internal read loop-back path from read-out port to read-in port.

        :param group: an index of a group which the read path belongs to.
        :return: None
        """
        self._css.activate_read_loop(group)

    def deactivate_read_loop(self, group: int) -> None:
        """disabling an internal read loop-back.

        :param group: an index of a group which the read path belongs to.
        :return: None
        """
        self._css.deactivate_read_loop(group)

    def is_loopedback_read(self, group: int) -> bool:
        """checking if an internal read loop-back path is activated or not.

        :param group: an index of a group which the read loop-back path belongs to.
        :return: True if the read loop-back path is activated.
        """
        return self._css.is_loopedback_read(group)

    def _dump_runit(
        self, group: int, rline: str, runit: int, rchannel_conf: Union[Dict[str, Any], None] = None
    ) -> Dict[str, Any]:
        if rchannel_conf is None:
            rchannel_conf = self.css.dump_rchannel(group, rline, self._get_rchannel_from_runit(group, rline, runit))
        else:
            rchannel_conf = copy.copy(rchannel_conf)

        # Notes: currently contents of runit_settings and rchannel_setttings are identical.
        runit_conf = rchannel_conf
        return runit_conf

    def dump_rfswitch(self, group: int, line: Union[int, str]) -> str:
        """dumpling the current configuration of a single RF switch

        :param group: a index of group the line belongs to.
        :param line: a group-local index of the line.
        :return:  the current configuration of the switch of the line.
        """
        state: bool = self._css.is_blocked_line(group, line)
        return self._decode_rfswitch_conf(group, line, state)

    def dump_rfswitches(self, exclude_subordinate: bool = True) -> Dict[Tuple[int, Union[int, str]], str]:
        """dumpling the current configurations of all the RF switches.

        :param exclude_subordinate: avoid to have copies of the same RF switches if True (default).
        :return: a mapping between a line and the configuration of its corresponding RF switch.
        """
        retval: Dict[Tuple[int, Union[int, str]], str] = {}
        for group, line in self._css.get_all_any_lines():
            if not (exclude_subordinate and self._css.is_subordinate_rfswitch(group, line)):
                retval[(group, line)] = self.dump_rfswitch(group, line)
        return retval

    def _dump_rline(self, group: int, rline: str) -> Dict[str, Any]:
        rl_conf = self._css.dump_rline(group, rline)
        rc_conf = rl_conf["channels"]
        del rl_conf["channels"]
        ru_conf: Dict[int, Any] = {}
        for runit in self.get_runits_of_rline(group, rline):
            ru_conf[runit] = self._dump_runit(
                group, rline, runit, rc_conf[self._get_rchannel_from_runit(group, rline, runit)]
            )
        rl_conf["runits"] = ru_conf
        return rl_conf

    def dump_line(self, group: int, line: Union[int, str]) -> Dict[str, Any]:
        """dumping the current configuration of a single line.

        :return: the current configuration of the line in dictionary.
        """
        if self.is_output_line(group, line):
            retval: Dict[str, Any] = self._css.dump_line(group, cast(int, line))
        elif self.is_input_line(group, line):
            retval = self._dump_rline(group, cast(str, line))
        else:
            raise ValueError(f"invalid (group, line): ({group}, {line})")
        return retval

    def dump_box(self) -> Dict[str, Dict[Union[int, Tuple[int, Union[int, str]]], Dict[str, Any]]]:
        """dumping the current configuration of the box.

        :return: the current configuration of the box in dictionary.
        """
        retval: Dict[str, Dict[Union[int, Tuple[int, Union[int, str]]], Dict[str, Any]]] = {"mxfes": {}, "lines": {}}

        for mxfe_idx in self.css.get_all_mxfes():
            retval["mxfes"][mxfe_idx] = {
                "channel_interporation_rate": self._css.get_channel_interpolation_rate(mxfe_idx),
                "main_interporation_rate": self._css.get_main_interpolation_rate(mxfe_idx),
            }

        for group in self.css.get_all_groups():
            for line in self._css.get_all_lines_of_group(group):
                retval["lines"][(group, line)] = self._css.dump_line(group, line)
            for rline in self._css.get_all_rlines_of_group(group):
                retval["lines"][(group, rline)] = self._dump_rline(group, rline)

        return retval

    def get_channels_of_line(self, group: int, line: int) -> Set[int]:
        if self.is_output_line(group, line):
            return set(range(self.css.get_num_channels_of_line(group, line)))
        else:
            raise ValueError(f"invalid output line: ({group}, {line})")

    def get_runits_of_rline(self, group: int, rline: str) -> Set[int]:
        # TODO: refine the implementation later
        if self.is_input_line(group, rline):
            num_capunt = 0
            for rch in range(self.css.get_num_rchannels_of_rline(group, rline)):
                capmod = self._get_capmod_from_rchannel(group, rline, rch)
                num_capunt += self.wss.num_capunit_of_capmod(capmod)
            return set(range(num_capunt))
        else:
            raise ValueError(f"invalid input line: ({group}, {rline})")

    def get_groups(self) -> set[int]:
        return self._css.get_all_groups()

    def get_output_lines(self) -> set[tuple[int, int]]:
        ol: set[tuple[int, int]] = set()
        for g in self._css.get_all_groups():
            for ln in self._css.get_all_lines_of_group(g):
                ol.add((g, ln))
        return ol

    def get_input_rlines(self) -> set[tuple[int, str]]:
        ol: set[tuple[int, str]] = set()
        for g in self._css.get_all_groups():
            for ln in self._css.get_all_rlines_of_group(g):
                ol.add((g, ln))
        return ol

    def get_names_of_wavedata(self, group: int, line: int, channel: int) -> set[str]:
        awgunit_idx = self._get_awg_from_channel(group, line, channel)
        return self.wss.get_names_of_wavedata(awgunit_idx)

    def register_wavedata(
        self,
        group: int,
        line: int,
        channel: int,
        name: str,
        iq: npt.NDArray[np.complex64],
        allow_update: bool = True,
        **kwdargs,
    ) -> None:
        awgunit_idx = self._get_awg_from_channel(group, line, channel)
        self.wss.register_wavedata(awgunit_idx, name, iq, allow_update=allow_update, **kwdargs)

    def has_wavedata(self, group: int, line: int, channel: int, name: str) -> bool:
        awgunit_idx = self._get_awg_from_channel(group, line, channel)
        return self.wss.has_wavedata(awgunit_idx, name)

    def delete_wavedata(self, group: int, line: int, channel: int, name: str) -> None:
        awgunit_idx = self._get_awg_from_channel(group, line, channel)
        self.wss.delete_wavedata(awgunit_idx, name)

    def initialize_all_awgunits(self) -> None:
        self.wss.initialize_all_awgunits()

    def initialize_all_capunits(self) -> None:
        self.wss.initialize_all_capunits()

    def get_current_timecounter(self) -> int:
        return self.wss.get_current_timecounter()

    def get_latest_sysref_timecounter(self) -> int:
        return self.wss.get_latest_sysref_timecounter()

    def get_averaged_sysref_offset(self, num_iteration: int = 100) -> float:
        return self.wss.get_averaged_sysref_offset(num_iteration)

    def start_wavegen(
        self,
        channels: Collection[Tuple[int, int, int]],
        timecounter: Optional[int] = None,
        *,
        timeout: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
        return_after_start_emission: Optional[bool] = None,
    ) -> AbstractStartAwgunitsTask:
        if len(channels) == 0:
            raise ValueError("no triggering channel are specified")
        for ch in channels:
            if not (isinstance(ch, tuple) and len(ch) == 3 and self.is_output_channel(*ch)):
                raise ValueError("invalid channel: {ch}")

        if timecounter and timecounter < 0:
            raise ValueError("negative timecounter is not allowed")
        if timeout and timeout <= 0.0:
            raise ValueError("non-positive timeout is not allowed")
        if polling_period and polling_period <= 0.0:
            raise ValueError("non-positive polling_period is not allowed")

        awgunit_idxs: Collection[int] = self._get_awgs_from_channels(channels)
        if timecounter is None:
            return self.wss.start_awgunits_now(
                awgunit_idxs,
                timeout=timeout,
                polling_period=polling_period,
                disable_timeout=disable_timeout,
                return_after_start_emission=return_after_start_emission or True,
            )
        else:
            return self.wss.start_awgunits_timed(
                awgunit_idxs,
                timecounter,
                timeout=timeout,
                polling_period=polling_period,
                disable_timeout=disable_timeout,
                return_after_start_emission=return_after_start_emission or False,
            )

    def start_capture_now(
        self,
        runits: Collection[Tuple[int, str, int]],
        *,
        timeout: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
    ) -> BoxIntrinsicStartCapunitsNowTask:
        if len(runits) == 0:
            raise ValueError("no capture units are specified")
        for ru in runits:
            if not (isinstance(ru, tuple) and len(ru) == 3 and self.is_input_runit(*ru)):
                raise ValueError("invalid runit: {ru}")

        if timeout and timeout <= 0.0:
            raise ValueError("non-positive timeout is not allowed")
        if polling_period and polling_period <= 0.0:
            raise ValueError("non-positive polling_period is not allowed")

        mapping: dict[tuple[int, int], tuple[int, str, int]] = {}
        for ru in runits:
            cu = self._get_capunit_from_runit(*ru)
            mapping[cu] = ru

        capunit_idxs = set(mapping.keys())
        return BoxIntrinsicStartCapunitsNowTask(
            self.wss.start_capunits_now(
                capunit_idxs, timeout=timeout, polling_period=polling_period, disable_timeout=disable_timeout
            ),
            mapping,
        )

    def start_capture_by_awg_trigger(
        self,
        runits: Collection[Tuple[int, str, int]],
        channels: Collection[Tuple[int, int, int]],
        timecounter: Optional[int] = None,
        *,
        timeout_before_trigger: Optional[float] = None,
        timeout_after_trigger: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
    ) -> tuple[BoxIntrinsicStartCapunitsByTriggerTask, AbstractStartAwgunitsTask]:
        if len(runits) == 0:
            raise ValueError("no capture units are specified")
        for ru in runits:
            if not (isinstance(ru, tuple) and len(ru) == 3 and self.is_input_runit(*ru)):
                raise ValueError("invalid runit: {ru}")
        if len(channels) == 0:
            raise ValueError("no triggering channel are specified")
        for ch in channels:
            if not (isinstance(ch, tuple) and len(ch) == 3 and self.is_output_channel(*ch)):
                raise ValueError("invalid channel: {ch}")
        if timecounter and timecounter < 0:
            raise ValueError("negative timecounter is not allowed")
        if timeout_before_trigger and timeout_before_trigger <= 0.0:
            raise ValueError("non-positive timeout_before_trigger is not allowed")
        if timeout_after_trigger and timeout_after_trigger <= 0.0:
            raise ValueError("non-positive timeout_after_trigger is not allowed")
        if polling_period and polling_period <= 0.0:
            raise ValueError("non-positive polling_period is not allowed")

        mapping: dict[tuple[int, int], tuple[int, str, int]] = {}
        capmod_idxs: set[int] = set()
        for ru in runits:
            cu = self._get_capunit_from_runit(*ru)
            mapping[cu] = ru
            capmod_idxs.add(cu[0])

        # Notes: any channel is fine since they all will start at the same time.
        trigger_idx = self._get_awg_from_channel(*list(channels)[0])
        for capmod_idx in capmod_idxs:
            self.wss.set_triggering_awg_to_line(capmod_idx, trigger_idx)

        if timecounter:
            cur = self.wss.get_current_timecounter()
            delta = self.wss.timecounter_to_second(timecounter - cur)
            if delta < StartAwgunitsTimedTask._TRIGGER_SETTABLE_MARGIN:
                raise RuntimeError(f"cannot schedule at the past or too close timecounter (= {timecounter})")
            if timeout_before_trigger is None:
                timeout_before_trigger = delta + StartAwgunitsTimedTask._START_TIMEOUT_MARGIN
        else:
            if timeout_before_trigger is None:
                timeout_before_trigger = StartAwgunitsNowTask._START_TIMEOUT

        capunit_idxs = set(mapping.keys())
        cap_task = BoxIntrinsicStartCapunitsByTriggerTask(
            self.wss.start_capunits_by_trigger(
                capunit_idxs,
                timeout_before_trigger=timeout_before_trigger,
                timeout_after_trigger=timeout_after_trigger,
                polling_period=polling_period,
                disable_timeout=disable_timeout,
            ),
            mapping,
        )

        gen_task: AbstractStartAwgunitsTask = self.start_wavegen(channels, timecounter, polling_period=polling_period)
        return cap_task, gen_task



import copy
import logging
import time
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any, Collection, Dict, Final, List, Mapping, Set, Tuple, Union

from quel_ic_config.ad5328 import Ad5328ConfigHelper
from quel_ic_config.ad9082 import LinkStatus, NcoFtw
from quel_ic_config.adrf6780 import Adrf6780ConfigHelper, Adrf6780LoSideband
from quel_ic_config.exstickge_sock_client import _ExstickgeProxyBase
from quel_ic_config.lmx2594 import Lmx2594ConfigHelper
from quel_ic_config.mixerboard_gpio import MixerboardGpioConfigHelper
from quel_ic_config.pathselectorboard_gpio import PathselectorboardGpioConfigHelper
from quel_ic_config.powerboard_pwm import PowerboardPwmConfigHelper
from quel_ic_config.quel_config_common import Quel1BoxType, Quel1RuntimeOption
from quel_ic_config.quel_ic import (
    Ad5328,
    Ad7490,
    Ad9082Generic,
    Adrf6780,
    Lmx2594,
    MixerboardGpio,
    PathselectorboardGpio,
    PowerboardPwm,
    QubeRfSwitchArray,
    Quel1TypeARfSwitchArray,
    Quel1TypeBRfSwitchArray,
)
from quel_ic_config.rfswitcharray import AbstractRfSwitchArrayMixin, RfSwitchArrayConfigHelper

logger = logging.getLogger(__name__)


class Quel1ConfigSubsystemBaseSlot(metaclass=ABCMeta):
    __slots__ = (
        "_css_addr",
        "_boxtype",
        "_runtime_options",
        "_proxy",
        "_ad9082",
        "_lmx2594",
        "_lmx2594_helper",
        "_adrf6780",
        "_adrf6780_helper",
        "_ad5328",
        "_ad5328_helper",
        "_ad7490",
        "_rfswitch",
        "_rfswitch_gpio_idx",
        "_rfswitch_helper",
        "_mixerboard_gpio",
        "_mixerboard_gpio_helper",
        "_pathselectorboard_gpio",
        "_pathselectorboard_gpio_helper",
        "_powerboard_pwm",
        "_powerboard_pwm_helper",
        "_tempctrl_watcher",
        "_tempctrl_auto_start_at_linkup",
    )

    _VALID_IC_NAME: Set[str] = {
        "ad9082",
        "lmx2594",
        "adrf6780",
        "ad5328",
        "ad7490",
        "gpio",
        "mixerboard_gpio",
        "pathselectorboard_gpio",
        "powerboard_pwm",
    }

    _DEFAULT_CONFIG_JSONFILE: str
    _NUM_IC: Dict[str, int]
    _GROUPS: Set[int]
    _MXFE_IDXS: Set[int]
    _DAC_IDX: Dict[Tuple[int, int], Tuple[int, int]]
    _ADC_IDX: Dict[Tuple[int, str], Tuple[int, int]]
    _ADC_CH_IDX: Dict[Tuple[int, str], Tuple[Tuple[int, int], ...]]

    def __init__(self):
        # variable defined in the importer class
        self._css_addr: str
        self._boxtype: Quel1BoxType
        self._runtime_options: Set[Quel1RuntimeOption]
        self._proxy: _ExstickgeProxyBase

        # TODO: types of IC proxy will be reconsidered when controler proxy (e.g. _ExstickgeProxyBase) is generalized.
        self._ad9082: Tuple[Ad9082Generic, ...] = ()
        self._lmx2594: Tuple[Lmx2594, ...] = ()
        self._lmx2594_helper: Tuple[Lmx2594ConfigHelper, ...] = ()
        self._adrf6780: Tuple[Adrf6780, ...] = ()
        self._adrf6780_helper: Tuple[Adrf6780ConfigHelper, ...] = ()
        self._ad5328: Tuple[Ad5328, ...] = ()
        self._ad5328_helper: Tuple[Ad5328ConfigHelper, ...] = ()
        self._rfswitch_gpio_idx: int = 0
        self._rfswitch: Union[AbstractRfSwitchArrayMixin, None] = None
        self._rfswitch_helper: Union[RfSwitchArrayConfigHelper, None] = None

    @property
    def has_lock(self) -> bool:
        return self._proxy.has_lock

    def get_num_ic(self, ic_name: str) -> int:
        if ic_name in self._VALID_IC_NAME:
            return self._NUM_IC.get(ic_name, 0)
        else:
            raise ValueError(f"invalid name of ic: '{ic_name}'")

    def _set_runtime_option(self, option: Quel1RuntimeOption, value: bool) -> None:
        if value:
            self._runtime_options.add(option)
        else:
            if option in self._runtime_options:
                self._runtime_options.remove(option)

    def _get_runtime_option(self, option: Quel1RuntimeOption) -> bool:
        return option in self._runtime_options

    def get_all_groups(self) -> Set[int]:
        return self._GROUPS

    def _validate_group(self, group: int) -> None:
        if group not in self._GROUPS:
            raise ValueError(f"an invalid group: {group}")

    def get_all_mxfes(self) -> Set[int]:
        return self._MXFE_IDXS

    def _validate_mxfe(self, mxfe_idx: int) -> None:
        if mxfe_idx not in self._MXFE_IDXS:
            raise ValueError(f"an invalid mxfe: {mxfe_idx}")

    def get_all_any_lines(self) -> Set[Tuple[int, Union[int, str]]]:
        any_lines: Set[Tuple[int, Union[int, str]]] = set()
        for g, l in self._DAC_IDX:
            any_lines.add((g, l))
        for g, rl in self._ADC_IDX:
            any_lines.add((g, rl))
        return any_lines

    def get_all_lines_of_group(self, group: int) -> Set[int]:
        self._validate_group(group)
        lines: Set[int] = set()
        for g, l in self._DAC_IDX:
            if g == group:
                lines.add(l)
        return lines

    def get_all_rlines_of_group(self, group: int) -> Set[str]:
        rlines: Set[str] = set()
        for g, l in self._ADC_IDX:
            if g == group:
                rlines.add(l)
        return rlines

    def _validate_line(self, group: int, line: int) -> None:
        if (group, line) not in self._DAC_IDX:
            raise ValueError(f"an invalid pair of group:{group} and line:{line}")

    def _validate_rline(self, group: int, rline: str) -> None:
        if (group, rline) not in self._ADC_IDX:
            raise ValueError(f"an invalid pair of group:{group} and rline:{rline}")

    def _validate_line_or_rline(self, group: int, line: Union[int, str]) -> None:
        if not ((group, line) in self._DAC_IDX or (group, line) in self._ADC_IDX):
            raise ValueError(f"an invalid pair of group:{group} and line/rline:{line}")

    @abstractmethod
    def get_num_channels_of_line(self, group: int, line: int):
        # a class managing MxFE should implement this.
        pass

    @abstractmethod
    def get_num_rchannels_of_rline(self, group: int, rline: str):
        # a class managing MxFE should implement this.
        pass

    def _validate_channel(self, group: int, line: int, channel: int) -> None:
        num_ch = self.get_num_channels_of_line(group, line)
        if not (isinstance(channel, int) and 0 <= channel < num_ch):
            raise ValueError(f"an invalid combination of group:{group}, line:{line}, and channel:{channel}")

    def _validate_rchannel(self, group: int, rline: str, rchannel: int) -> None:
        num_rch = self.get_num_rchannels_of_rline(group, rline)
        if not (isinstance(rchannel, int) and 0 <= rchannel < num_rch):
            raise ValueError(f"an invalid combination of group:{group}, rline:{rline}, and rchannel:{rchannel}")


class Quel1ConfigSubsystemRoot(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = 0,
        timeout: float = 0.0,
        sender_limit_by_binding: bool = False,
    ):
        super(Quel1ConfigSubsystemRoot, self).__init__()
        self._css_addr: str = css_addr
        self._boxtype: Quel1BoxType = boxtype
        self._runtime_options: Set[Quel1RuntimeOption] = set()
        self._proxy: _ExstickgeProxyBase = self._create_exstickge_proxy(port, timeout, sender_limit_by_binding)

    def __del__(self):
        if hasattr(self, "_proxy"):
            self._proxy.terminate()

    def initialize(self) -> None:
        pass

    def get_default_config_filename(self) -> Path:
        return Path(self._DEFAULT_CONFIG_JSONFILE)

    def get_num_ics(self) -> Dict[str, int]:
        return copy.copy(self._NUM_IC)

    @abstractmethod
    def _create_exstickge_proxy(self, port: int, timeout: float, sender_limit_by_binding: bool) -> _ExstickgeProxyBase:
        pass

    @property
    def boxtype(self) -> Quel1BoxType:
        # Notes: at the css layer, boxtype is Quel1BoxType.
        #        at the box layer, boxtype is its alias in string.
        return self._boxtype

    @property
    def ipaddr_css(self) -> str:
        return self._css_addr

    @abstractmethod
    def configure_peripherals(
        self,
        param: Dict[str, Any],
        *,
        ignore_access_failure_of_adrf6780: Union[Collection[int], None] = None,
        ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None,
    ) -> None:
        """configure ICs other than MxFEs and PLLs for their reference clock.

        :param param: configuration parameters.
        :param ignore_access_failure_of_adrf6780: a collection of index of ADRF6780 to ignore access failure during the
                                                  initialization process.
        :param ignore_lock_failure_of_lmx2594: a collection of index of LMX2594 to ignore PLL lock failure during the
                                               initialization process.
        :return: None
        """
        pass

    @abstractmethod
    def configure_all_mxfe_clocks(
        self, param: Dict[str, Any], *, ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None
    ) -> None:
        """configure PLLs for clocks of MxFEs.

        :param param: configuration parameters.
        :param ignore_lock_failure_of_lmx2594: a collection of index of LMX2594 to ignore PLL lock failure during the
                                               initialization process.
        :return: None
        """
        pass

    @abstractmethod
    def configure_mxfe(
        self,
        mxfe_idx: int,
        param: dict[str, Any],
        *,
        hard_reset: bool,
        soft_reset: bool,
        use_204b: bool,
        use_bg_cal: bool,
        ignore_crc_error: bool,
    ) -> bool:
        """configure an MxFE and its related data objects. PLLs for their clock must be set up in advance.

        :param mxfe_idx: an index of a group which the target MxFE belongs to.
        :param param: configuration parameters.
        :param hard_reset: enabling hard reset of the MxFE before the initialization if available.
        :param soft_reset: enabling soft reset of the MxFE before the initialization.
        :param use_204b: using a workaround method to link the MxFE up during the initialization if True.
        :param ignore_crc_error: ignoring CRC error flag at the validation of the link status if True.
        :return: True if the target MxFE is available.
        """
        pass

    @abstractmethod
    def reconnect_mxfe(
        self,
        mxfe_idx: int,
        *,
        ignore_crc_error: bool,
    ) -> bool:
        """configure an MxFE and its related data objects. PLLs for their clock must be set up in advance.

        :param mxfe_idx: an index of a group which the target MxFE belongs to.
        :param ignore_crc_error: ignoring CRC error flag at the validation of the link status if True.
        :return: True if the target MxFE is available.
        """
        pass


class Quel1GenericConfigSubsystemAd9082Mixin(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    @abstractmethod
    def _construct_ad9082(self): ...

    def validate_chip_id(self, mxfe_idx: int):
        self._validate_mxfe(mxfe_idx)
        chip_id = self.ad9082[mxfe_idx].device_chip_id_get()
        if chip_id.prod_id != 0x9082 or chip_id.dev_revision < 3:
            raise RuntimeError(
                f"unexpected chip_id of mxfe-#{mxfe_idx}: prod_id = {chip_id.prod_id:04x}, rev = {chip_id.dev_revision}"
            )

    @property
    def allow_dual_modulus_nco(self) -> bool:
        return self._get_runtime_option(Quel1RuntimeOption.ALLOW_DUAL_MODULUS_NCO)

    @allow_dual_modulus_nco.setter
    def allow_dual_modulus_nco(self, v: bool) -> None:
        self._set_runtime_option(Quel1RuntimeOption.ALLOW_DUAL_MODULUS_NCO, v)

    @property
    def ad9082(self) -> Tuple[Ad9082Generic, ...]:
        return self._ad9082

    def check_link_status(
        self, mxfe_idx: int, mxfe_init: bool = False, ignore_crc_error: bool = False
    ) -> Tuple[bool, int, str]:
        link_status, crc_error = self.ad9082[mxfe_idx].get_link_status()
        link_status_strs = ["("]
        # Notes: be aware that assuming that any elements of LinkStatus start with "LINK_STATUS_".
        # TODO: should define str() for LinkStatus.
        link_status_strs.append(f"link status is {str(link_status).split('.')[-1][12:]}, ")
        if crc_error == 0:
            link_status_strs.append("No CRC error")
        else:
            link_status_strs.append("CRC error is detected")
        link_status_strs.append(")")
        link_status_str = "".join(link_status_strs)
        mxfe_name = f"{self._css_addr}:AD9082-#{mxfe_idx}"
        judge: bool = False
        if link_status == LinkStatus.LINK_STATUS_LOCKED:
            if crc_error == 0:
                judge = True
            elif crc_error == 1 and ignore_crc_error:
                judge = True

        if judge:
            if crc_error == 0:
                lglv: int = logging.INFO
                if mxfe_init:
                    diag: str = f"{mxfe_name} links up successfully {link_status_str}"
                else:
                    diag = f"{mxfe_name} has linked up healthy {link_status_str}"
            else:
                lglv = logging.WARNING
                if mxfe_init:
                    diag = f"{mxfe_name} links up successfully with ignored crc error {link_status_str}"
                else:
                    diag = f"{mxfe_name} has linked up but crc error is detected thereafter {link_status_str}"
        else:
            lglv = logging.WARNING
            if mxfe_init:
                diag = f"{mxfe_name} fails to link up {link_status_str}"
            else:
                diag = f"{mxfe_name} has not linked up yet {link_status_str}"

        return judge, lglv, diag

    def clear_crc_error(self, mxfe_idx: int) -> None:
        self.ad9082[mxfe_idx].clear_crc_error()

    def _validate_frequency_info(
        self, mxfe_idx: int, freq_type: str, freq_in_hz: Union[float, None], ftw: Union[NcoFtw, None]
    ) -> Tuple[float, NcoFtw]:
        # Notes: Assuming that group is already validated.
        # Notes: decimation rates can be different among DACs, however,
        #        QuEL-1 doesn't allow such configuration currently.
        if (freq_in_hz is not None and ftw is not None) or (freq_in_hz is None and ftw is None):
            raise ValueError("either of freq_in_hz or ftw is required")
        elif freq_in_hz is not None:
            # TODO: consider to use calc_*_*_ftw_rational().
            if freq_type == "dac_cnco":
                ftw = self.ad9082[mxfe_idx].calc_dac_cnco_ftw(freq_in_hz, fractional_mode=self.allow_dual_modulus_nco)
            elif freq_type == "dac_fnco":
                ftw = self.ad9082[mxfe_idx].calc_dac_fnco_ftw(freq_in_hz, fractional_mode=self.allow_dual_modulus_nco)
            elif freq_type == "adc_cnco":
                ftw = self.ad9082[mxfe_idx].calc_adc_cnco_ftw(freq_in_hz, fractional_mode=self.allow_dual_modulus_nco)
            elif freq_type == "adc_fnco":
                ftw = self.ad9082[mxfe_idx].calc_adc_fnco_ftw(freq_in_hz, fractional_mode=self.allow_dual_modulus_nco)
            else:
                raise ValueError(f"invalid freq_type: {freq_in_hz}")
        elif ftw is not None:
            if not isinstance(ftw, NcoFtw):
                raise TypeError("unexpected ftw is given")
            # Notes: freq_in_hz is computed back from ftw for logger messages.
            if freq_type == "dac_cnco":
                freq_in_hz = self.ad9082[mxfe_idx].calc_dac_cnco_freq(ftw)
            elif freq_type == "dac_fnco":
                freq_in_hz = self.ad9082[mxfe_idx].calc_dac_fnco_freq(ftw)
            elif freq_type == "adc_cnco":
                raise NotImplementedError
            elif freq_type == "adc_fnco":
                raise NotImplementedError
            else:
                raise ValueError(f"invalid freq_type: {freq_type}")
        else:
            raise AssertionError  # never happens

        return freq_in_hz, ftw

    def get_dac_idx(self, group: int, line: int) -> Tuple[int, int]:
        self._validate_line(group, line)
        return self._DAC_IDX[(group, line)]

    def get_fduc_idx(self, group: int, line: int, channel: int) -> Tuple[int, int]:
        self._validate_channel(group, line, channel)
        mxfe_idx, dac_idx = self.get_dac_idx(group, line)
        return mxfe_idx, self.ad9082[mxfe_idx].get_fduc_of_dac(dac_idx)[channel]

    def get_num_channels_of_line(self, group: int, line: int) -> int:
        mxfe_idx, dac_idx = self.get_dac_idx(group, line)
        return len(self.ad9082[mxfe_idx].get_fduc_of_dac(dac_idx))

    def get_adc_idx(self, group: int, rline: str) -> Tuple[int, int]:
        self._validate_rline(group, rline)
        return self._ADC_IDX[(group, rline)]

    def get_rline_from_adc_idx(self, mxfe_idx: int, adc_idx: int) -> Union[Tuple[int, str], None]:
        self._validate_mxfe(mxfe_idx)
        for gl, (m, c) in self._ADC_IDX.items():
            if m == mxfe_idx and c == adc_idx:
                return gl
        else:
            return None

    def get_num_rchannels_of_rline(self, group: int, rline: str) -> int:
        self._validate_rline(group, rline)
        return len(self._ADC_CH_IDX[group, rline])

    def get_fddc_idx(self, group: int, rline: str, rchannel: int) -> Tuple[int, int]:
        self._validate_rchannel(group, rline, rchannel)
        return self._ADC_CH_IDX[group, rline][rchannel]

    def set_dac_cnco(
        self, group: int, line: int, freq_in_hz: Union[float, None] = None, ftw: Union[NcoFtw, None] = None
    ) -> None:
        """setting CNCO frequency of a transmitter line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :param freq_in_hz: the CNCO frequency of the line in Hz.
        :param ftw: an FTW can be passed instead of freq_in_hz if necessary.
        :return: None
        """
        mxfe_idx, dac_idx = self.get_dac_idx(group, line)
        freq_in_hz_, ftw_ = self._validate_frequency_info(mxfe_idx, "dac_cnco", freq_in_hz, ftw)
        logger.info(f"DAC-CNCO{dac_idx} of MxFE{mxfe_idx} of {self._css_addr} is set to {freq_in_hz_}Hz (ftw = {ftw_})")
        self.ad9082[mxfe_idx].set_dac_cnco({dac_idx}, ftw_)

    def get_dac_cnco(self, group: int, line: int) -> float:
        """getting CNCO frequency of a transmitter line.

        :param group: an index of group which the line belongs to.
        :param line: a group-local index of the line.
        :return: the current CNCO frequency of the line in Hz.
        """
        mxfe_idx, dac_idx = self.get_dac_idx(group, line)
        ftw = self.ad9082[mxfe_idx].get_dac_cnco(dac_idx)
        return self.ad9082[mxfe_idx].calc_dac_cnco_freq(ftw)

    def is_equivalent_dac_cnco(self, group: int, line: int, freq0: float, freq1: float) -> bool:
        mxfe_idx, _ = self.get_dac_idx(group, line)
        return self.ad9082[mxfe_idx].is_equivalent_dac_cnco(freq0, freq1, self.allow_dual_modulus_nco)

    def set_dac_fnco(
        self,
        group: int,
        line: int,
        channel: int,
        freq_in_hz: Union[float, None] = None,
        ftw: Union[NcoFtw, None] = None,
    ) -> None:
        """setting FNCO frequency of the transmitter channel.

        :param group: an index of a group which the channel belongs to.
        :param line: a group-local index of a line which the channel belongs to.
        :param channel: a line-local index of the chennel.
        :param freq_in_hz: the FNCO frequency of the channel in Hz.
        :param ftw: an FTW can be passed instead of freq_in_hz if necessary.
        :return: None
        """
        mxfe_idx, fnco_idx = self.get_fduc_idx(group, line, channel)
        freq_in_hz_, ftw_ = self._validate_frequency_info(mxfe_idx, "dac_fnco", freq_in_hz, ftw)
        logger.info(
            f"DAC-FNCO{fnco_idx} of MxFE{mxfe_idx} of {self._css_addr} is set to {freq_in_hz_}Hz (ftw = {ftw_})"
        )
        self.ad9082[mxfe_idx].set_dac_fnco({fnco_idx}, ftw_)

    def get_dac_fnco(self, group: int, line: int, channel: int) -> float:
        """getting FNCO frequency of a transmitter channel.

        :param group: an index of a group which the channel belongs to.
        :param line: a group-local index of a line which the channel belongs to.
        :param channel: a line-local index of the chennel.
        :return: the current FNCO frequency of the channel in Hz.
        """
        mxfe_idx, fnco_idx = self.get_fduc_idx(group, line, channel)
        ftw = self.ad9082[mxfe_idx].get_dac_fnco(fnco_idx)
        return self.ad9082[mxfe_idx].calc_dac_fnco_freq(ftw)

    def is_equivalent_dac_fnco(self, group: int, line: int, freq0: float, freq1: float) -> bool:
        mxfe_idx, _ = self.get_dac_idx(group, line)
        return self.ad9082[mxfe_idx].is_equivalent_dac_fnco(freq0, freq1, self.allow_dual_modulus_nco)

    def set_adc_cnco(
        self, group: int, rline: str, freq_in_hz: Union[float, None] = None, ftw: Union[NcoFtw, None] = None
    ):
        """setting CNCO frequency of a receiver line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :param freq_in_hz: the CNCO frequency of the line in Hz.
        :param ftw: an FTW can be passed instead of freq_in_hz if necessary.
        :return: None
        """
        mxfe_idx, adc_idx = self.get_adc_idx(group, rline)
        freq_in_hz_, ftw_ = self._validate_frequency_info(mxfe_idx, "adc_cnco", freq_in_hz, ftw)
        logger.info(f"ADC-CNCO{adc_idx} of MxFE{mxfe_idx} of {self._css_addr} is set to {freq_in_hz_}Hz (ftw = {ftw_})")
        self.ad9082[mxfe_idx].set_adc_cnco({adc_idx}, ftw_)

    def get_adc_cnco(self, group: int, rline: str) -> float:
        """getting CNCO frequency of a receiver line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: the current CNCO frequency of the line in Hz.
        """
        mxfe_idx, adc_idx = self.get_adc_idx(group, rline)
        ftw = self.ad9082[mxfe_idx].get_adc_cnco(adc_idx)
        return self.ad9082[mxfe_idx].calc_adc_cnco_freq(ftw)

    def is_equivalent_adc_cnco(self, group: int, rline: str, freq0: float, freq1: float) -> bool:
        mxfe_idx, _ = self.get_adc_idx(group, rline)
        return self.ad9082[mxfe_idx].is_equivalent_adc_cnco(freq0, freq1, self.allow_dual_modulus_nco)

    def set_adc_fnco(
        self,
        group: int,
        rline: str,
        rchannel: int,
        freq_in_hz: Union[float, None] = None,
        ftw: Union[NcoFtw, None] = None,
    ):
        """setting FNCO frequency of a receiver channel.

        :param group: an index of a group which the channel belongs to.
        :param line: a group-local index of a line which the channel belongs to.
        :param channel: a line-local index of the chennel.
        :param freq_in_hz: the FNCO frequency of the channel in Hz.
        :param ftw: an FTW can be passed instead of freq_in_hz if necessary.
        :return: None
        """
        mxfe_idx, fnco_idx = self.get_fddc_idx(group, rline, rchannel)
        freq_in_hz_, ftw_ = self._validate_frequency_info(mxfe_idx, "adc_fnco", freq_in_hz, ftw)
        logger.info(
            f"ADC-FNCO{fnco_idx} of MxFE{mxfe_idx} of {self._css_addr} is set to {freq_in_hz_}Hz (ftw = {ftw_})"
        )
        self.ad9082[mxfe_idx].set_adc_fnco({fnco_idx}, ftw_)

    def get_adc_fnco(self, group: int, rline: str, rchannel: int) -> float:
        """getting FNCO frequency of a receiver channel.

        :param group: an index of a group which the channel belongs to.
        :param line: a group-local index of a line which the channel belongs to.
        :param channel: a line-local index of the chennel.
        :return: the current FNCO frequency of the channel in Hz.
        """
        mxfe_idx, fnco_idx = self.get_fddc_idx(group, rline, rchannel)
        ftw = self.ad9082[mxfe_idx].get_adc_fnco(fnco_idx)
        return self.ad9082[mxfe_idx].calc_adc_fnco_freq(ftw)

    def is_equivalent_adc_fnco(self, group: int, rline: str, freq0: float, freq1: float) -> bool:
        mxfe_idx, _ = self.get_adc_idx(group, rline)
        return self.ad9082[mxfe_idx].is_equivalent_adc_fnco(freq0, freq1, self.allow_dual_modulus_nco)

    def set_pair_cnco(self, group_dac: int, line_dac: int, group_adc: int, rline_adc: str, freq_in_hz: float) -> None:
        dac_mxfe_idx, dac_idx = self.get_dac_idx(group_dac, line_dac)
        adc_mxfe_idx, adc_idx = self.get_adc_idx(group_adc, rline_adc)

        dac_clk = self.ad9082[dac_mxfe_idx].device.dev_info.dac_freq_hz
        adc_clk = self.ad9082[adc_mxfe_idx].device.dev_info.adc_freq_hz
        # TODO: relax constrains by-need basis.
        if dac_clk % adc_clk != 0:
            raise RuntimeError(f"ratio of dac_clk (= {dac_clk}Hz) and adc_clk (= {adc_clk}Hz) is not N:1")
        ratio: int = dac_clk // adc_clk

        freq_in_hz_, dac_ftw = self._validate_frequency_info(dac_mxfe_idx, "dac_cnco", freq_in_hz, None)
        adc_ftw = dac_ftw.multiply(ratio)
        logger.info(
            f"DAC-CNCO{dac_idx} of MxFE{dac_mxfe_idx} and ADC-CNCO{adc_idx} of MxFE{adc_mxfe_idx} of {self._css_addr} "
            f"are set to {freq_in_hz_}Hz (dac_ftw = {dac_ftw}, adc_ftw = {adc_ftw})"
        )
        self.ad9082[dac_mxfe_idx].set_dac_cnco({dac_idx}, dac_ftw)
        self.ad9082[adc_mxfe_idx].set_adc_cnco({adc_idx}, adc_ftw)

    def get_link_status(self, mxfe_idx: int) -> Tuple[LinkStatus, int]:
        """getting the status of the datalink between a MxFE in a group and the FPGA.

        :param mxfe_idx: an index of the group which the target MxFE belongs to.
        :return: the content of the registers of the MxFE at the addresses of 0x55E and 0x5BB.
        """
        self._validate_mxfe(mxfe_idx)
        return self._ad9082[mxfe_idx].get_link_status()

    def set_fullscale_current(self, group: int, line: int, fsc: int) -> None:
        """setting the full-scale current of a DAC of a line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line
        :param fsc: full-scale current of the DAC of the line in uA.
        :return: None
        """
        mxfe_idx, dac_idx = self.get_dac_idx(group, line)
        self.ad9082[mxfe_idx].set_fullscale_current(1 << dac_idx, fsc)

    def get_fullscale_current(self, group: int, line: int) -> int:
        """geting the current full scale current of a DAC of a line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line
        :return: the current full-scale current of the DAC in uA
        """
        mxfe_idx, dac_idx = self.get_dac_idx(group, line)
        return self.ad9082[mxfe_idx].get_fullscale_current(dac_idx)

    def is_equal_fullscale_current(self, group: int, line: int, fsc0: int, fsc1: int) -> bool:
        mxfe_idx, _ = self.get_dac_idx(group, line)
        return self.ad9082[mxfe_idx].is_equal_fullscale_current(fsc0, fsc1)

    def get_main_interpolation_rate(self, mxfe_idx: int) -> int:
        """getting the current main interpolation rate of a MxFE of a group.

        :param mxfe_idx: an index of the group which the target MxFE belongs to.
        :return: main interpolation rate
        """
        self._validate_mxfe(mxfe_idx)
        return self.ad9082[mxfe_idx].get_main_interpolation_rate()

    def get_channel_interpolation_rate(self, mxfe_idx: int) -> int:
        """getting the current channel interpolation rate of a MxFE of a group.

        :param mxfe_idx: a group which the target MxFE belongs to.
        :return: channel interpolation rate
        """
        self._validate_mxfe(mxfe_idx)
        return self.ad9082[mxfe_idx].get_channel_interpolation_rate()

    def get_virtual_adc_select(self, mxfe_idx: int) -> List[int]:
        """getting converter select matrix of ADC virtual converters after FDDC.

        :param mxfe_idx: a group which the target MxFE belongs to.
        :return: converter select matrix in list
        """
        self._validate_mxfe(mxfe_idx)
        return self.ad9082[mxfe_idx].get_virtual_adc_select()

    def get_crc_error_counts(self, mxfe_idx: int) -> List[int]:
        return self.ad9082[mxfe_idx].get_crc_error_counts()

    def get_mxfe_temperature_range(self, mxfe_idx: int) -> Tuple[int, int]:
        """getting the current die temperatures of a MxFE of a group.

        :param mxfe_idx: an index of the group which the target MxFE belongs to.
        :return: a pair of maximum and minimum temperatures in the die.
        """
        self._validate_mxfe(mxfe_idx)
        temp_max, temp_min = self.ad9082[mxfe_idx].get_temperatures()
        return temp_max, temp_min


class Quel1ConfigSubsystemLmx2594Mixin(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    _LO_IDX: Dict[Tuple[int, Union[int, str]], Tuple[int, int]]  # (group, line) --> (idx_of_ic, idx_of_output)

    def _construct_lmx2594(self):
        self._lmx2594: Tuple[Lmx2594, ...] = tuple(Lmx2594(self._proxy, idx) for idx in range(self._NUM_IC["lmx2594"]))
        self._lmx2594_helper: Tuple[Lmx2594ConfigHelper, ...] = tuple(Lmx2594ConfigHelper(ic) for ic in self._lmx2594)

    @property
    def lmx2594(self) -> Tuple[Lmx2594, ...]:
        return self._lmx2594

    @property
    def lmx2594_helper(self) -> Tuple[Lmx2594ConfigHelper, ...]:
        return self._lmx2594_helper

    def init_lmx2594(
        self,
        idx: int,
        param: Mapping[str, Mapping[str, Mapping[str, Union[int, bool]]]],
        ignore_lock_failure: bool = False,
    ) -> bool:
        # Notes: ignore_lock_failure is prepared for experimental purposes.
        ic, helper = self._lmx2594[idx], self._lmx2594_helper[idx]
        ic.soft_reset()
        for name, fields in param["registers"].items():
            helper.write_field(name, **fields)
        helper.flush()
        is_locked = ic.calibrate()
        if not is_locked:
            if ignore_lock_failure:
                logger.error(
                    f"failed to lock PLL of {self._css_addr}:LMX2594-#{idx}, keep running with an unlocked PLL"
                )
            else:
                raise RuntimeError(f"failed to lock PLL of {self._css_addr}:LMX2594-#{idx}")
        return is_locked

    def set_lo_multiplier(self, group: int, line: Union[int, str], freq_multiplier: int) -> bool:
        """setting the frequency multiplier of a PLL of a line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line which the PLL belongs to.
        :param freq_multiplier: the frequency multiplier to set.
        :return: True if the frequency calibration is completed successfully.
        """
        self._validate_line_or_rline(group, line)
        if (group, line) not in self._LO_IDX:
            raise ValueError(f"no LO is available for (group:{group}, line:{line})")
        lo_idx, _ = self._LO_IDX[(group, line)]
        ic = self._lmx2594[lo_idx]
        ic.set_lo_multiplier(freq_multiplier)
        logger.info(f"updating LO frequency[{lo_idx}] of {self._css_addr} to {freq_multiplier * 100}MHz")
        return ic.calibrate()

    def get_lo_multiplier(self, group: int, line: Union[int, str]) -> int:
        """get the current frequency multiplier of a PLL of a line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line which the PLL belongs to.
        :return: the current frequency multiplier of the PLL.
        """
        self._validate_line_or_rline(group, line)
        if (group, line) not in self._LO_IDX:
            raise ValueError(f"no LO is available for (group:{group}, line:{line})")
        lo_idx, _ = self._LO_IDX[(group, line)]
        return self._lmx2594[lo_idx].get_lo_multiplier()

    def set_divider_ratio(self, group: int, line: Union[int, str], divide_ratio: int) -> None:
        """setting the output frequency divide ration of a line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line which the PLL belongs to.
        :param divide_ratio: a ratio of the frequency output divider.
        """
        self._validate_line_or_rline(group, line)
        if (group, line) not in self._LO_IDX:
            raise ValueError(f"no LO is available for (group:{group}, line:{line})")
        if divide_ratio < 1:
            raise ValueError(f"invalid divide ratio {divide_ratio} for (group:{group}, line:{line})")

        lo_idx, outpin = self._LO_IDX[(group, line)]
        ic = self._lmx2594[lo_idx]
        current_ratios = ic.get_divider_ratio()
        modified_ratios = list(current_ratios)
        # Notes: both output pin should generate the same freq (an unused outpin is kept as is.)
        for i, ratio in enumerate(current_ratios):
            if ratio != 0:
                modified_ratios[i] = divide_ratio
        ic.set_divider_ratio(*modified_ratios)
        logger.info(
            f"updating frequency divide ratio of (group:{group}, line:{line}) of {self._css_addr} to {divide_ratio}"
        )

    def get_divider_ratio(self, group: int, line: Union[int, str]) -> int:
        """get the current output frequency divide ratio of a line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line which the PLL belongs to.
        :return: the current frequency divide ratio of the output
        """
        self._validate_line_or_rline(group, line)
        if (group, line) not in self._LO_IDX:
            raise ValueError(f"no LO is available for group:{group}, line:{line}")
        lo_idx, outpin = self._LO_IDX[(group, line)]

        ic = self._lmx2594[lo_idx]
        if outpin not in {0, 1}:
            raise AssertionError(f"invalid outpin '{outpin}' is specified in class definition")
        return ic.get_divider_ratio()[outpin]


class Quel1ConfigSubsystemAd6780Mixin(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    _MIXER_IDX: Dict[Tuple[int, int], int]

    def _construct_adrf6780(self):
        self._adrf6780: Tuple[Adrf6780, ...] = tuple(
            Adrf6780(self._proxy, idx) for idx in range(self._NUM_IC.get("adrf6780", 0))
        )
        self._adrf6780_helper: Tuple[Adrf6780ConfigHelper, ...] = tuple(
            Adrf6780ConfigHelper(ic) for ic in self._adrf6780
        )

    @property
    def adrf6780(self) -> Tuple[Adrf6780, ...]:
        return self._adrf6780

    @property
    def adrf6780_helper(self) -> Tuple[Adrf6780ConfigHelper, ...]:
        return self._adrf6780_helper

    def init_adrf6780(
        self, idx, param: Mapping[str, Mapping[str, Mapping[str, Union[int, bool]]]], ignore_id_mismatch: bool = False
    ) -> bool:
        # Notes: ignore_id_mismatch is a workaround for some ill-designed boards.
        ic, helper = self._adrf6780[idx], self._adrf6780_helper[idx]
        ic.soft_reset(parity_enable=False)
        matched, (chip_id, chip_rev) = ic.check_id()
        if matched:
            logger.info(f"chip revision of ADRF6780[{idx}] of {self._css_addr} is ({chip_id}, {chip_rev})")
        else:
            if ignore_id_mismatch:
                logger.error(
                    f"unexpected chip revision of ADRF6780[{idx}] of {self._css_addr}: ({chip_id}, {chip_rev}), "
                    f"keep running with an unstable mixer"
                )
            else:
                raise RuntimeError(
                    f"unexpected chip revision of ADRF6780[{idx}] of {self._css_addr}: ({chip_id}, {chip_rev})"
                )
        for name, fields in param["registers"].items():
            helper.write_field(name, **fields)
        helper.flush()
        return matched

    def _validate_sideband(self, sideband: str):
        if not isinstance(sideband, str):
            raise TypeError(f"unexpected sideband: {sideband}")
        if sideband not in {"L", "U"}:
            raise ValueError(f"unexpected sideband: {sideband}")

    def _get_mixer_idx(self, group, line):
        if (group, line) in self._MIXER_IDX:
            return self._MIXER_IDX[(group, line)]
        raise ValueError(f"no configurable mixer is available for group:{group}, line:{line}")

    def set_sideband(self, group: int, line: int, sideband: str) -> None:
        """setting the active sideband of a mixer of a line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line which the mixer belongs to.
        :param sideband: the chosen active sideband, either of "U" for upper or "L" for lower.
        :return: None
        """
        mixer_idx = self._get_mixer_idx(group, line)
        self._validate_sideband(sideband)
        if sideband == "L":
            logger.info(f"ADRF6780[{mixer_idx}] of {self._css_addr} is set to LSB mode")
            ic = self._adrf6780[mixer_idx]
            ic.set_lo_sideband(Adrf6780LoSideband.Lsb)
        elif sideband == "U":
            logger.info(f"ADRF6780[{mixer_idx}] of {self._css_addr} is set to USB mode")
            ic = self._adrf6780[mixer_idx]
            ic.set_lo_sideband(Adrf6780LoSideband.Usb)
        else:
            raise AssertionError  # never happens

    def get_sideband(self, group: int, line: int) -> str:
        """getting the active sideband of a mixer of a line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line which the mixer belongs to.
        :return: the current sideband of the mixer.
        """
        mixer_idx = self._get_mixer_idx(group, line)
        if mixer_idx is None:
            raise ValueError(f"no configurable mixer is available for group:{group}, line:{line}")

        ic = self._adrf6780[mixer_idx]
        ssb = ic.get_lo_sideband()
        if ssb == Adrf6780LoSideband.Usb:
            return "U"
        elif ssb == Adrf6780LoSideband.Lsb:
            return "L"
        else:
            raise AssertionError


class Quel1ConfigSubsystemAd5328Mixin(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    MAX_VATT: Final[int] = 0xC9B

    _VATT_IDX: Dict[Tuple[int, int], Tuple[int, int]]

    def _construct_ad5328(self):
        self._ad5328: Tuple[Ad5328, ...] = tuple(
            Ad5328(self._proxy, idx) for idx in range(self._NUM_IC.get("ad5328", 0))
        )
        self._ad5328_helper: Tuple[Ad5328ConfigHelper, ...] = tuple(Ad5328ConfigHelper(ic) for ic in self._ad5328)

    @property
    def ad5328(self) -> Tuple[Ad5328, ...]:
        return self._ad5328

    @property
    def ad5328_helper(self) -> Tuple[Ad5328ConfigHelper, ...]:
        return self._ad5328_helper

    def init_ad5328(self, idx: int, param: Mapping[str, Mapping[str, Mapping[str, Union[int, bool]]]]) -> None:
        ic, helper = self._ad5328[idx], self._ad5328_helper[idx]
        for name, fields in param["registers"].items():
            helper.write_field(name, **fields)
        helper.flush()
        ic.update_dac()

    def _validate_vatt(self, vatt: int) -> int:
        if not (0 <= vatt < 4096):
            raise ValueError("invalid vatt value")
        if vatt > self.MAX_VATT:
            vatt = self.MAX_VATT
            logger.warning(f"vatt is clipped to {self.MAX_VATT}")
        return vatt

    def _get_vatt_idx(self, group: int, line: int):
        if (group, line) in self._VATT_IDX:
            return self._VATT_IDX[(group, line)]
        else:
            raise ValueError(f"no variable attenuator is available for group:{group}, line:{line}")

    def set_vatt(self, group: int, line: int, vatt: int) -> None:
        """setting the VATT control voltage of a mixer of a line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line which the mixer belongs to.
        :param vatt: a control voltage of the VATT (in unit of 3.3V / 4096, see data sheet of ADRF6780 for details).
        :return: None
        """
        ic_idx, vatt_idx = self._get_vatt_idx(group, line)
        vatt = self._validate_vatt(vatt)
        logger.info(f"VATT{vatt_idx} of {self._css_addr} is set to 0x{vatt:03x}")
        ic = self._ad5328[ic_idx]
        ic.set_output(vatt_idx, vatt, update_dac=True)

    def get_vatt_carboncopy(self, group: int, line: int) -> Union[int, None]:
        """getting the VATT control voltage of a mixer of a line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line which the mixer belongs to.
        :return: the latest copy of the control voltage of the VATT of the line passed to set_vatt().
        """
        ic_idx, vatt_idx = self._get_vatt_idx(group, line)
        ic = self._ad5328[ic_idx]
        return ic.get_output_carboncopy(vatt_idx)


class NoRfSwitchError(Exception):
    pass


class NoLoopbackPathError(Exception):
    pass


class Quel1ConfigSubsystemNoRfswitch(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    def is_subordinate_rfswitch(self, group: int, line: Union[int, str]) -> bool:
        return False

    def pass_line(self, group: int, line: Union[int, str]) -> None:
        logger.info(f"making (group:{group}, line:{line}) passing")

    def block_line(self, group: int, line: Union[int, str]) -> None:
        raise NoRfSwitchError()

    def is_blocked_line(self, group: int, line: Union[int, str]) -> bool:
        return False

    def is_passed_line(self, group: int, line: Union[int, str]) -> bool:
        return True

    def activate_monitor_loop(self, group: int) -> None:
        raise NoLoopbackPathError()

    def deactivate_monitor_loop(self, group: int) -> None:
        logger.info(f"deactivate monitor loop (group:{group})")

    def is_loopedback_monitor(self, group: int) -> bool:
        return False

    def activate_read_loop(self, group: int) -> None:
        raise NoLoopbackPathError()

    def deactivate_read_loop(self, group: int) -> None:
        logger.info(f"deactivate read loop (group:{group})")

    def is_loopedback_read(self, group: int) -> bool:
        return False


class Quel1ConfigSubsystemRfswitch(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    _RFSWITCH_NAME: Dict[Tuple[int, Union[int, str]], Tuple[int, str]]
    _RFSWITCH_SUBORDINATE_OF: Dict[Tuple[int, Union[int, str]], Tuple[int, Union[int, str]]]

    # TODO: this is not good with respect to the hierarchical structure.
    def _construct_rfswitch(self, gpio_idx: int):
        if self._boxtype in {Quel1BoxType.QuEL1_TypeA, Quel1BoxType.QuEL1_NTT}:
            gpio_tmp: AbstractRfSwitchArrayMixin = Quel1TypeARfSwitchArray(self._proxy, gpio_idx)
        elif self._boxtype == Quel1BoxType.QuEL1_TypeB:
            gpio_tmp = Quel1TypeBRfSwitchArray(self._proxy, gpio_idx)
        elif self._boxtype in {
            Quel1BoxType.QuBE_RIKEN_TypeA,
            Quel1BoxType.QuBE_RIKEN_TypeB,
        }:
            gpio_tmp = QubeRfSwitchArray(self._proxy, gpio_idx)
        else:
            raise ValueError(f"invalid boxtype: {self._boxtype}")
        self._rfswitch_gpio_idx = gpio_idx
        self._rfswitch: AbstractRfSwitchArrayMixin = gpio_tmp
        self._rfswitch_helper: RfSwitchArrayConfigHelper = RfSwitchArrayConfigHelper(self._rfswitch)

    @property
    def rfswitch(self) -> AbstractRfSwitchArrayMixin:
        return self._rfswitch

    @property
    def rfswitch_helper(self) -> RfSwitchArrayConfigHelper:
        return self._rfswitch_helper

    def init_rfswitch(self, param: Dict[str, Dict[str, Dict[str, Union[int, bool]]]]) -> None:
        # XXX: = self._param["gpio"][self._rfswitch_gpio_idx]
        helper = self._rfswitch_helper
        for name, fields in param["registers"].items():
            helper.write_field(name, **fields)
        helper.flush()

    def _alternate_rfswitch(self, group: int, **sw_update: bool) -> None:
        # Note: assuming that group is already validated.
        ic, helper = self._rfswitch, self._rfswitch_helper
        current_sw: int = ic.read_reg(0)
        helper.write_field(f"Group{group}", **sw_update)
        helper.flush()
        altered_sw: int = ic.read_reg(0)
        time.sleep(0.01)
        logger.debug(
            f"alter the state of RF switch array of {self._css_addr} from {current_sw:014b} to {altered_sw:014b}"
        )

    def _get_rfswitch_name(self, group: int, line: Union[int, str]) -> Tuple[int, str]:
        if (group, line) in self._RFSWITCH_NAME:
            return self._RFSWITCH_NAME[(group, line)]
        else:
            raise ValueError(f"no switch available for group:{group}, line:{line}")

    def is_subordinate_rfswitch(self, group: int, line: Union[int, str]) -> bool:
        return (group, line) in self._RFSWITCH_SUBORDINATE_OF

    def pass_line(self, group: int, line: Union[int, str]) -> None:
        """allowing a line to emit RF signal from its corresponding SMA connector.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: None
        """
        logger.info(f"making (group:{group}, line:{line}) passing")
        swgroup, swname = self._get_rfswitch_name(group, line)
        self._alternate_rfswitch(swgroup, **{swname: False})

    def block_line(self, group: int, line: Union[int, str]) -> None:
        """blocking a line to emit RF signal from its corresponding SMA connector.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: None
        """
        logger.info(f"making (group:{group}, line:{line}) blocked")
        swgroup, swname = self._get_rfswitch_name(group, line)
        self._alternate_rfswitch(swgroup, **{swname: True})

    def is_blocked_line(self, group: int, line: Union[int, str]) -> bool:
        """checking if the emission of RF signal of a line is blocked or not.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: True if the RF signal emission from its corresponding SMA connector is blocked.
        """
        swgroup, swname = self._get_rfswitch_name(group, line)
        return getattr(self._rfswitch_helper.read_reg(swgroup), swname)

    def is_passed_line(self, group: int, line: Union[int, str]) -> bool:
        """checking if the emission of RF signal from a line is passed or not.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: True if the RF signal emission from its corresponding SMA connector is passed.
        """
        return not self.is_blocked_line(group, line)

    def activate_monitor_loop(self, group: int) -> None:
        """enabling an internal loop-back of a monitor path of a group, from monitor-out to monitor-in.

        :param group: an index of the group.
        :return: None
        """
        logger.info(f"activate monitor loop (group:{group})")
        swgroup, swname = self._get_rfswitch_name(group, "m")
        self._alternate_rfswitch(swgroup, **{swname: True})

    def deactivate_monitor_loop(self, group: int) -> None:
        """disabling an internal loop-back of a monitor path of a group.

        :param group: an index of the group.
        :return: None
        """
        logger.info(f"deactivate monitor loop (group:{group})")
        swgroup, swname = self._get_rfswitch_name(group, "m")
        self._alternate_rfswitch(swgroup, **{swname: False})

    def is_loopedback_monitor(self, group: int) -> bool:
        """checking if a monitor loop-back path of a group is activated or not.

        :param group: an index of the group.
        :return: True if the monitor loop-back path of the group is activated.
        """
        swgroup, swname = self._get_rfswitch_name(group, "m")
        return getattr(self._rfswitch_helper.read_reg(swgroup), swname)

    def activate_read_loop(self, group: int) -> None:
        """enabling an internal loop-back of a read path of a group, from monitor-out to monitor-in.

        :param group: an index of the group.
        :return: None
        """
        logger.info(f"activate read loop (group:{group})")
        swgroup, swname = self._get_rfswitch_name(group, "r")
        self._alternate_rfswitch(swgroup, **{swname: True})

    def deactivate_read_loop(self, group: int) -> None:
        """disabling an internal loop-back of a read path of a group.

        :param group: an index of the group.
        :return: None
        """
        logger.info(f"deactivate read loop (group:{group})")
        swgroup, swname = self._get_rfswitch_name(group, "r")
        self._alternate_rfswitch(swgroup, **{swname: False})

    def is_loopedback_read(self, group: int) -> bool:
        """checking if a read loop-back path of a group is activated or not.

        :param group: an index of the group.
        :return: True if the read loop-back path of the group is activated.
        """
        swgroup, swname = self._get_rfswitch_name(group, "r")
        return getattr(self._rfswitch_helper.read_reg(swgroup), swname)


class Quel1ConfigSubsystemMixerboardGpioMixin(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    def _construct_mixerboard_gpio(self):
        self._mixerboard_gpio: Tuple[MixerboardGpio, ...] = tuple(
            MixerboardGpio(self._proxy, idx) for idx in range(self._NUM_IC.get("mixerboard_gpio", 0))
        )

        self._mixerboard_gpio_helper: Tuple[MixerboardGpioConfigHelper, ...] = tuple(
            MixerboardGpioConfigHelper(ic) for ic in self._mixerboard_gpio
        )

    @property
    def mixerboard_gpio(self) -> Tuple[MixerboardGpio, ...]:
        return self._mixerboard_gpio

    @property
    def mixerboard_gpio_helper(self) -> Tuple[MixerboardGpioConfigHelper, ...]:
        return self._mixerboard_gpio_helper

    def init_mixerboard_gpio(self, idx: int, param: Mapping[str, Mapping[str, Mapping[str, Union[int, bool]]]]) -> None:
        # XXX: = self._param["mixerboard_gpio"][idx]
        helper = self.mixerboard_gpio_helper[idx]
        for name, fields in param["registers"].items():
            helper.write_field(name, **fields)
        helper.flush()


# TODO: reconsider the design, this looks too wet.
class Quel1ConfigSubsystemPathselectorboardGpioMixin(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    _RFSWITCH_IDX: Dict[Tuple[int, Union[int, str]], Tuple[int, int]]
    _RFSWITCH_SUBORDINATE_OF: Dict[Tuple[int, Union[int, str]], Tuple[int, Union[int, str]]]

    def _construct_pathselectorboard_gpio(self):
        self._pathselectorboard_gpio: Tuple[PathselectorboardGpio, ...] = tuple(
            PathselectorboardGpio(self._proxy, idx) for idx in range(self._NUM_IC.get("pathselectorboard_gpio", 0))
        )

        self._pathselectorboard_gpio_helper: Tuple[PathselectorboardGpioConfigHelper, ...] = tuple(
            PathselectorboardGpioConfigHelper(ic) for ic in self._pathselectorboard_gpio
        )

    @property
    def pathselectorboard_gpio(self) -> Tuple[PathselectorboardGpio, ...]:
        return self._pathselectorboard_gpio

    @property
    def pathselectorboard_gpio_helper(self) -> Tuple[PathselectorboardGpioConfigHelper, ...]:
        return self._pathselectorboard_gpio_helper

    def init_pathselectorboard_gpio(
        self, idx: int, param: Mapping[str, Mapping[str, Mapping[str, Union[int, bool]]]]
    ) -> None:
        # XXX: = self._param["pathselectorboard_gpio"][idx]
        helper = self.pathselectorboard_gpio_helper[idx]
        for name, fields in param["registers"].items():
            helper.write_field(name, **fields)
        helper.flush()

    def is_subordinate_rfswitch(self, group: int, line: Union[int, str]) -> bool:
        return (group, line) in self._RFSWITCH_SUBORDINATE_OF

    def _alternate_rfswitch(self, idx: int, **sw_update: bool) -> None:
        # Note: assuming that group is already validated.
        ic, helper = self._pathselectorboard_gpio, self._pathselectorboard_gpio_helper
        current_sw: int = ic[idx].read_reg(0)
        helper[idx].write_field(0, **sw_update)
        helper[idx].flush()
        altered_sw: int = ic[idx].read_reg(0)
        time.sleep(0.01)
        logger.debug(
            f"alter the state of RF switch array {idx} of {self._css_addr} from {current_sw:06b} to {altered_sw:06b}"
        )

    def _get_rfswitch_name(self, group: int, line: Union[int, str]) -> Tuple[int, str]:
        if (group, line) in self._RFSWITCH_IDX:
            idx, bit = self._RFSWITCH_IDX[(group, line)]
            return idx, f"b{bit:02}"
        else:
            raise ValueError(f"no switch available for group:{group}, line:{line}")

    def pass_line(self, group: int, line: Union[int, str]) -> None:
        """allowing a line to emit RF signal from its corresponding SMA connector.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: None
        """
        logger.info(f"making (group:{group}, line:{line}) passing")
        swgroup, swname = self._get_rfswitch_name(group, line)
        self._alternate_rfswitch(swgroup, **{swname: False})

    def block_line(self, group: int, line: Union[int, str]) -> None:
        """blocking a line to emit RF signal from its corresponding SMA connector.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: None
        """
        logger.info(f"making (group:{group}, line:{line}) blocked")
        swgroup, swname = self._get_rfswitch_name(group, line)
        self._alternate_rfswitch(swgroup, **{swname: True})

    def is_blocked_line(self, group: int, line: Union[int, str]) -> bool:
        """checking if the emission of RF signal of a line is blocked or not.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: True if the RF signal emission from its corresponding SMA connector is blocked.
        """
        swgroup, swname = self._get_rfswitch_name(group, line)
        return getattr(self._pathselectorboard_gpio_helper[swgroup].read_reg(0), swname)

    def is_passed_line(self, group: int, line: Union[int, str]) -> bool:
        """checking if the emission of RF signal from a line is passed or not.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: True if the RF signal emission from its corresponding SMA connector is passed.
        """
        return not self.is_blocked_line(group, line)

    def activate_monitor_loop(self, group: int) -> None:
        """enabling an internal loop-back of a monitor path of a group, from monitor-out to monitor-in.

        :param group: an index of the group.
        :return: None
        """
        logger.info(f"activate monitor loop (group:{group})")
        swgroup, swname = self._get_rfswitch_name(group, "m")
        self._alternate_rfswitch(swgroup, **{swname: True})

    def deactivate_monitor_loop(self, group: int) -> None:
        """disabling an internal loop-back of a monitor path of a group.

        :param group: an index of the group.
        :return: None
        """
        logger.info(f"deactivate monitor loop (group:{group})")
        swgroup, swname = self._get_rfswitch_name(group, "m")
        self._alternate_rfswitch(swgroup, **{swname: False})

    def is_loopedback_monitor(self, group: int) -> bool:
        """checking if a monitor loop-back path of a group is activated or not.

        :param group: an index of the group.
        :return: True if the monitor loop-back path of the group is activated.
        """
        swgroup, swname = self._get_rfswitch_name(group, "m")
        return getattr(self._pathselectorboard_gpio_helper[swgroup].read_reg(0), swname)

    def activate_read_loop(self, group: int) -> None:
        """enabling an internal loop-back of a read path of a group, from monitor-out to monitor-in.

        :param group: an index of the group.
        :return: None
        """
        logger.info(f"activate read loop (group:{group})")
        swgroup, swname = self._get_rfswitch_name(group, "r")
        self._alternate_rfswitch(swgroup, **{swname: True})

    def deactivate_read_loop(self, group: int) -> None:
        """disabling an internal loop-back of a read path of a group.

        :param group: an index of the group.
        :return: None
        """
        logger.info(f"deactivate read loop (group:{group})")
        swgroup, swname = self._get_rfswitch_name(group, "r")
        self._alternate_rfswitch(swgroup, **{swname: False})

    def is_loopedback_read(self, group: int) -> bool:
        """checking if a read loop-back path of a group is activated or not.

        :param group: an index of the group.
        :return: True if the read loop-back path of the group is activated.
        """
        swgroup, swname = self._get_rfswitch_name(group, "r")
        return getattr(self._pathselectorboard_gpio_helper[swgroup].read_reg(0), swname)


class Quel1ConfigSubsystemAd7490Mixin(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    def _construct_ad7490(self):
        self._ad7490: Tuple[Ad7490, ...] = tuple(
            Ad7490(self._proxy, idx) for idx in range(self._NUM_IC.get("ad7490", 0))
        )

    @property
    def ad7490(self) -> Tuple[Ad7490, ...]:
        return self._ad7490

    def init_ad7490(self, idx: int, param: Mapping[str, Mapping[str, Mapping[str, Union[int, bool]]]]) -> None:
        # XXX: = self._param["ad7490"][idx]
        self._ad7490[idx].set_default_config(**param["registers"]["Config"])
        # Notes: default_config is not applied to the IC now because it'll be applied just before reading channels.


class Quel1ConfigSubsystemPowerboardPwmMixin(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    def _construct_powerboard_pwm(self):
        self._powerboard_pwm: Tuple[PowerboardPwm, ...] = tuple(
            PowerboardPwm(self._proxy, idx) for idx in range(self._NUM_IC.get("powerboard_pwm", 0))
        )

        self._powerboard_pwm_helper: Tuple[PowerboardPwmConfigHelper, ...] = tuple(
            PowerboardPwmConfigHelper(ic) for ic in self._powerboard_pwm
        )

    @property
    def powerboard_pwm(self) -> Tuple[PowerboardPwm, ...]:
        return self._powerboard_pwm

    @property
    def powerboard_pwm_helper(self) -> Tuple[PowerboardPwmConfigHelper, ...]:
        return self._powerboard_pwm_helper



from collections.abc import Collection, Sequence
from concurrent.futures import Future
from pathlib import Path
from typing import Any, Protocol, Union, runtime_checkable

from quel_ic_config.ad9082 import LinkStatus, NcoFtw
from quel_ic_config.exstickge_coap_tempctrl_client import Quel1seTempctrlState
from quel_ic_config.quel1_thermistor import Quel1Thermistor
from quel_ic_config.quel_config_common import Quel1BoxType

Quel1AnyLineType = Union[int, str]


@runtime_checkable
class Quel1AnyConfigSubsystem(Protocol):
    def initialize(self) -> None: ...

    # configuration info
    def get_default_config_filename(self) -> Path: ...

    def get_num_ics(self) -> dict[str, int]: ...

    # introspecting box-info
    @property
    def boxtype(self) -> Quel1BoxType: ...

    @property
    def ipaddr_css(self) -> str: ...

    @property
    def has_lock(self) -> bool: ...

    # introspecting group-line-channel
    def get_all_groups(self) -> set[int]: ...

    def get_all_any_lines(self) -> set[tuple[int, Quel1AnyLineType]]: ...

    def get_all_lines_of_group(self, group: int) -> set[int]: ...

    def get_all_rlines_of_group(self, group: int) -> set[str]: ...

    def get_num_channels_of_line(self, group: int, line: int) -> int: ...

    def get_num_rchannels_of_rline(self, group: int, rline: str) -> int: ...

    # introspecting hardware/firmware info
    def get_num_ic(self, ic_name: str) -> int: ...

    def get_all_mxfes(self) -> set[int]: ...

    def get_rline_from_adc_idx(self, mxfe_idx: int, adc_idx: int) -> Union[tuple[int, str], None]: ...

    # low-level APIs for controlling ICs
    def validate_chip_id(self, mxfe_idx: int) -> bool: ...

    @property
    def allow_dual_modulus_nco(self) -> bool: ...

    @allow_dual_modulus_nco.setter
    def allow_dual_modulus_nco(self, v: bool) -> None: ...

    def get_link_status(self, mxfe_idx: int) -> tuple[LinkStatus, int]: ...

    def get_crc_error_counts(self, mxfe_idx: int) -> list[int]: ...

    def get_main_interpolation_rate(self, mxfe_idx: int) -> int: ...

    def get_channel_interpolation_rate(self, mxfe_idx: int) -> int: ...

    def get_dac_idx(self, group: int, line: int) -> tuple[int, int]: ...

    def get_fduc_idx(self, group: int, line: int, channel: int) -> tuple[int, int]: ...

    def get_adc_idx(self, group: int, rline: str) -> tuple[int, int]: ...

    def get_fddc_idx(self, group: int, rline: str, rchannel: int) -> tuple[int, int]: ...

    def get_virtual_adc_select(self, mxfe_idx: int) -> list[int]: ...

    def set_fullscale_current(self, group: int, line: int, fsc: int) -> None: ...

    def get_fullscale_current(self, group: int, line: int) -> int: ...

    def is_equal_fullscale_current(self, group: int, line: int, fsc0: int, fsc1: int) -> bool: ...

    def set_dac_cnco(
        self, group: int, line: int, freq_in_hz: Union[float, None] = None, ftw: Union[NcoFtw, None] = None
    ) -> None: ...

    def get_dac_cnco(self, group: int, line: int) -> float: ...

    def is_equivalent_dac_cnco(self, group: int, line: int, freq0: float, freq1: float) -> bool: ...

    def set_dac_fnco(
        self,
        group: int,
        line: int,
        channel: int,
        freq_in_hz: Union[float, None] = None,
        ftw: Union[NcoFtw, None] = None,
    ) -> None: ...

    def get_dac_fnco(self, group: int, line: int, channel: int) -> float: ...

    def is_equivalent_dac_fnco(self, group: int, line: int, freq0: float, freq1: float) -> bool: ...

    def set_adc_cnco(
        self, group: int, rline: str, freq_in_hz: Union[float, None] = None, ftw: Union[NcoFtw, None] = None
    ): ...

    def get_adc_cnco(self, group: int, rline: str) -> float: ...

    def is_equivalent_adc_cnco(self, group: int, rline: str, freq0: float, freq1: float) -> bool: ...

    def set_adc_fnco(
        self,
        group: int,
        rline: str,
        rchannel: int,
        freq_in_hz: Union[float, None] = None,
        ftw: Union[NcoFtw, None] = None,
    ): ...

    def get_adc_fnco(self, group: int, rline: str, rchannel: int) -> float: ...

    def is_equivalent_adc_fnco(self, group: int, rline: str, freq0: float, freq1: float) -> bool: ...

    def set_pair_cnco(
        self, group_dac: int, line_dac: int, group_adc: int, rline_adc: str, freq_in_hz: float
    ) -> None: ...

    def get_lo_multiplier(self, group: int, line: Quel1AnyLineType) -> int: ...

    def set_lo_multiplier(self, group: int, line: Quel1AnyLineType, freq_multiplier: int) -> bool: ...

    def get_divider_ratio(self, group: int, line: Quel1AnyLineType) -> int: ...

    def set_divider_ratio(self, group: int, line: Quel1AnyLineType, divide_ratio: int) -> None: ...

    # def get_vatt(self, group: int, line: int) -> int: ...

    def set_vatt(self, group: int, line: int, vatt: int) -> None: ...

    def get_sideband(self, group: int, line: int) -> str: ...

    def set_sideband(self, group: int, line: int, sideband: str) -> None: ...

    def pass_line(self, group: int, line: Quel1AnyLineType) -> None: ...

    def is_passed_line(self, group: int, line: Quel1AnyLineType) -> bool: ...

    def block_line(self, group: int, line: Quel1AnyLineType) -> None: ...

    def is_blocked_line(self, group: int, line: Quel1AnyLineType) -> bool: ...

    def activate_monitor_loop(self, group: int) -> None: ...

    def deactivate_monitor_loop(self, group: int) -> None: ...

    def is_loopedback_monitor(self, group: int) -> bool: ...

    def activate_read_loop(self, group: int) -> None: ...

    def deactivate_read_loop(self, group: int) -> None: ...

    def is_loopedback_read(self, group: int) -> bool: ...

    def is_subordinate_rfswitch(self, group: int, line: Union[int, str]) -> bool: ...

    # high-level APIs
    def configure_all_mxfe_clocks(
        self, param: dict[str, Any], *, ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None
    ) -> None: ...

    def configure_mxfe(
        self,
        mxfe_idx: int,
        param: dict[str, Any],
        *,
        hard_reset: bool = False,
        soft_reset: bool = False,
        use_204b: bool = True,
        use_bg_cal: bool = False,
        ignore_crc_error: bool = False,
    ) -> bool: ...

    def configure_peripherals(
        self,
        param: dict[str, Any],
        *,
        ignore_access_failure_of_adrf6780: Union[Collection[int], None] = None,
        ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None,
    ) -> None: ...

    def reconnect_mxfe(
        self,
        mxfe_idx: int,
        *,
        ignore_crc_error: bool = False,
    ) -> bool: ...

    def check_link_status(
        self, mxfe_idx: int, mxfe_init: bool = False, ignore_crc_error: bool = False
    ) -> tuple[bool, int, str]: ...

    def clear_crc_error(self, mxfe_idx: int) -> None: ...

    def dump_channel(self, group: int, line: int, channel: int) -> dict[str, Any]: ...

    def dump_line(self, group: int, line: int) -> dict[str, Any]: ...

    def dump_rchannel(self, group: int, rline: str, rchannel: int) -> dict[str, Any]: ...

    def dump_rline(self, group: int, rline: str) -> dict[str, Any]: ...

    # high-level APIs for temperature control
    @property
    def tempctrl_auto_start_at_linkup(self) -> bool: ...

    def get_tempctrl_state(self) -> Quel1seTempctrlState: ...

    def get_tempctrl_state_count(self) -> int: ...

    def start_tempctrl(self, new_count: Union[int, None] = None) -> None: ...

    def get_mxfe_temperature_range(self, mxfe_idx: int) -> tuple[int, int]: ...


@runtime_checkable
class Quel1seAnyConfigSubsystem(Quel1AnyConfigSubsystem, Protocol):
    @classmethod
    def get_thermistor_desc(cls) -> dict[tuple[int, int], Quel1Thermistor]: ...

    @classmethod
    def get_actuator_desc(cls) -> dict[str, tuple[str, int]]: ...

    def get_tempctrl_temperature(self) -> Future[dict[str, float]]: ...

    def get_tempctrl_actuator_output(self) -> dict[str, dict[str, float]]: ...

    def get_tempctrl_setpoint(self) -> dict[str, list[float]]: ...

    def set_tempctrl_setpoint(self, fan: Sequence[float], heater: Sequence[float]): ...



import logging
from collections.abc import Collection

from . import _config

logger = logging.getLogger(__name__)

_CLOCK_FREQ = 125_000_000
_TRIGGER_GRID_STEP_DEFAULT = 16

BoxName = str


def _round_up_to_grid(val: float, grid_step: int) -> int:
    return int(val + (grid_step - 1)) // grid_step * grid_step


class StableCountProposer:
    def __init__(self, trigger_grid_step: int = _TRIGGER_GRID_STEP_DEFAULT):
        self._name_to_offset: dict[str, int] = {}
        self._trigger_grid_step: int = trigger_grid_step

    def get_offset(self, box_name) -> int:
        return self._name_to_offset[box_name]

    def set_offset(self, box_name: str, offset: int):
        self._name_to_offset[box_name] = offset

    def propose_trigger_counts(
        self,
        current_count: int,
        target_names: Collection[str],
        delay_sec: float = 0.15,
    ) -> dict[BoxName, int]:
        reference = _round_up_to_grid(current_count + _CLOCK_FREQ * delay_sec, self._trigger_grid_step)
        return {name: reference + self._name_to_offset.get(name, 0) for name in target_names}

    @classmethod
    def from_deskew_configuration(
        cls, deskew_conf: _config.DeskewConfiguration, trigger_grid_step: int = _TRIGGER_GRID_STEP_DEFAULT
    ):
        inst = cls(trigger_grid_step=trigger_grid_step)
        for box in deskew_conf.boxes:
            inst.set_offset(box.name, box.timecounter_offset)
        return inst



import logging
from typing import Final

import numpy as np
import numpy.typing as npt
from e7awghal.capdata import CapIqDataReader

import quel_ic_config as qi

from . import _config

logger = logging.getLogger(__name__)

_SAMPLING_RATE = 500e6
_SAMPLES_PER_WORD = 4

_BLANK_WAVEDATA_NAME = "__quel_deskew_tools_blank"
_BLANK_WAVEDATA_LENGTH = 64
_BLANK_WAVEDATA = np.zeros((_BLANK_WAVEDATA_LENGTH,), dtype=np.complex64)
_DUMMY_CAPSECTION_NAME = "__quel_deskew_tools_dummy_capsection"
_AWG_INIT_BLANK_REFERENCE_WORD_DEFAULT = 1280

BoxName = str


def register_blank_wavedata(box: qi.Quel1Box, port, channel):
    box.register_wavedata(port, channel, _BLANK_WAVEDATA_NAME, _BLANK_WAVEDATA)


class E7awgDelayCompensator:
    """Compensates delays by adjusting parameters for e7awg."""

    _MINIMUM_BLANK_IN_LAST_CAP_SECTION: Final[int] = 21
    _MINIMUM_BLANK_IN_LAST_CHUNK: Final[int] = (_BLANK_WAVEDATA_LENGTH // 4) + 15

    def __init__(self, awg_init_blank_reference_word: int = _AWG_INIT_BLANK_REFERENCE_WORD_DEFAULT):
        """Initializes the E7awgParameterAdjuster.

        :param awg_init_blank_reference_word: The reference number of blank words
            at the beginning of the AWG sequence. This value serves as a baseline
            for all subsequent adjustments.
        """
        self._awg_init_blank_reference_word = awg_init_blank_reference_word

    def adjust_awg_param(self, awg_param: qi.AwgParam, init_blank_offset_word: int) -> qi.AwgParam:
        """Adjusts AwgParam by inserting a blank chunk at the beginning.

        This method calculates the necessary blank word count based on a reference
        and an offset, then inserts a `WaveChunk` with this blank count to the
        front of the AWG sequence.

        :param awg_param: The ``AwgParam`` object to be adjusted.
        :param init_blank_offset_word: The offset value to be added to the reference
                                       blank word count.
        :returns: The modified ``AwgParam`` object with the blank chunk inserted.
        """
        copied = awg_param.model_copy(deep=True)
        if self._check_if_awg_param_already_adjusted(awg_param):
            raise ValueError("AwgParam is already adjusted.")

        if len(copied.chunks) > 15:
            raise ValueError("The number of chunks must be less than 16.")

        init_blank_word = self._awg_init_blank_reference_word + init_blank_offset_word - (_BLANK_WAVEDATA_LENGTH // 4)
        num_wait_word_div_by_16, num_blank_word_of_dummy_chunk = divmod(init_blank_word, 16)
        copied.num_wait_word += 16 * num_wait_word_div_by_16

        blank_chunk = qi.WaveChunk(
            name_of_wavedata=_BLANK_WAVEDATA_NAME, num_repeat=1, num_blank_word=num_blank_word_of_dummy_chunk
        )
        copied.chunks.insert(0, blank_chunk)

        if copied.num_repeat > 1:
            last_chunk = copied.chunks[-1]
            if last_chunk.num_repeat > 1:
                raise ValueError(
                    "When num_repeat of AwgParam > 1, num_repeat of the last chunk must be 1 for adjustment."
                )
            if last_chunk.num_blank_word < self._MINIMUM_BLANK_IN_LAST_CHUNK:
                raise ValueError(
                    f"When num_repeat > 1, the number of num_blank_word in the last chunk must be equal to or greater than {self._MINIMUM_BLANK_IN_LAST_CHUNK} for adjustment."
                )
            last_chunk.num_blank_word -= _BLANK_WAVEDATA_LENGTH // 4
            last_chunk.num_blank_word -= num_blank_word_of_dummy_chunk

        return copied

    def _check_if_awg_param_already_adjusted(self, awg_param: qi.AwgParam):
        return awg_param.chunks[0].name_of_wavedata == _BLANK_WAVEDATA_NAME

    def adjust_cap_param(self, cap_param: qi.CapParam, init_blank_offset_word: int) -> qi.CapParam:
        """Adjusts CapParam.

        This method inserts a dummy section for timing adjustment at the beginning without
        changing the total length of all sections combined.

        :param cap_param: The ``CapParam`` object to be adjusted.
        :param init_blank_offset_word: The offset value used to calculate the total
            blank words for the CAP sequence.
        :returns: The modified ``CapParam`` object.
        :raises ValueError: If the number of ``num_blank_word`` in the last section is not
            greater than 20, which is required for the adjustment logic.
        """
        copied = cap_param.model_copy(deep=True)
        if self._check_if_cap_param_already_adjusted(cap_param):
            raise ValueError("CapParam is already adjusted.")

        copied.sections.insert(0, qi.CapSection(name=_DUMMY_CAPSECTION_NAME, num_capture_word=4, num_blank_word=1))

        cap_init_blank_word = self._awg_init_blank_reference_word + init_blank_offset_word - (4 + 1)
        num_wait_word_div_by_16, num_blank_word_of_dummy_section = divmod(cap_init_blank_word, 16)
        copied.num_wait_word += 16 * num_wait_word_div_by_16
        copied.sections[0].num_blank_word += num_blank_word_of_dummy_section
        if copied.num_repeat > 1:
            last_section = copied.sections[-1]
            if last_section.num_blank_word < self._MINIMUM_BLANK_IN_LAST_CAP_SECTION:
                raise ValueError(
                    f"When num_repeat > 1, the number of num_blank_word in the last section must be equal to or greater than {self._MINIMUM_BLANK_IN_LAST_CAP_SECTION} for adjustment."
                )
            last_section.num_blank_word -= 4 + 1
            last_section.num_blank_word -= num_blank_word_of_dummy_section

        return copied

    def _check_if_cap_param_already_adjusted(self, cap_param: qi.CapParam):
        return cap_param.sections[0].name == _DUMMY_CAPSECTION_NAME

    def get_minimum_blank_in_last_cap_section(self):
        return self._MINIMUM_BLANK_IN_LAST_CAP_SECTION

    def get_minimum_blank_in_last_chunk(self):
        return self._MINIMUM_BLANK_IN_LAST_CHUNK


def _picosec_to_word(ps: int) -> int:
    word_index = (ps * (1e-12 * _SAMPLING_RATE)) / _SAMPLES_PER_WORD
    return int(round(word_index))


class WaitAmountResolver:
    def __init__(self):
        self._name_to_wait_word: dict[str, int] = {}
        self._name_port_to_wait_word_offset: dict[tuple[str, qi.Quel1PortType], int] = {}

    def register_wait_word(self, box_name: BoxName, wait_word: int):
        self._name_to_wait_word[box_name] = wait_word

    def register_wait_word_port_offset(self, box_name: BoxName, port: qi.Quel1PortType, wait_word: int):
        self._name_port_to_wait_word_offset[(box_name, port)] = wait_word

    @classmethod
    def from_deskew_configuration(cls, deskew_conf: _config.DeskewConfiguration):
        inst = cls()
        for box in deskew_conf.boxes:
            inst.register_wait_word(box.name, _picosec_to_word(box.wait_ps))
            for port in box.ports:
                inst.register_wait_word_port_offset(box.name, port.port, _picosec_to_word(port.wait_ps_offset))
        return inst

    def get_word_to_wait(self, box_name: str, port: qi.Quel1PortType) -> int:
        if (box_name, port) in self._name_port_to_wait_word_offset:
            return self._name_to_wait_word.get(box_name, 0) + self._name_port_to_wait_word_offset[(box_name, port)]
        else:
            return self._name_to_wait_word.get(box_name, 0)


def extract_wave_dict(iq_reader: CapIqDataReader) -> dict[str, npt.NDArray[np.complex64]]:
    d = iq_reader.as_wave_dict()
    d.pop(_DUMMY_CAPSECTION_NAME, None)
    return d


def extract_wave_list(iq_reader: CapIqDataReader) -> list[npt.NDArray[np.complex64]]:
    l = iq_reader.as_wave_list()
    return l[1:]



from pathlib import Path
from typing import Union

import matplotlib.pyplot as plt
import numpy as np


def plot_iqs(
    iq_dict,
    *,
    t_offset: int = 0,
    same_range: bool = True,
    show_graph: bool = True,
    output_filename: Union[Path, None] = None,
) -> None:
    n_plot = len(iq_dict)

    m = 0
    if same_range:
        for _, iq in iq_dict.items():
            m = max(m, np.max(np.abs(iq)))

    fig, axs = plt.subplots(n_plot, sharex="col")
    if n_plot == 1:
        axs = [axs]
    fig.set_size_inches(10.0, 2.0 * n_plot)
    fig.subplots_adjust(bottom=max(0.025, 0.125 / n_plot), top=min(0.975, 1.0 - 0.05 / n_plot))
    for idx, (title, iq) in enumerate(iq_dict.items()):
        if not same_range:
            m = np.max(np.abs(iq))
        t = np.arange(0, len(iq)) - t_offset
        axs[idx].plot(t, np.real(iq))
        axs[idx].plot(t, np.imag(iq))
        axs[idx].set_ylim((-m * 1.1, m * 1.1))
        axs[idx].text(0.05, 0.1, title, transform=axs[idx].transAxes)
    if output_filename is not None:
        plt.savefig(output_filename)
    if show_graph:
        plt.show()
